#!/usr/bin/env python3

import sys
import os
import argparse
sys.path.append(os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    "lib"
))
from tgs.exporters import export_svg
from tgs.parsers.tgs import parse_tgs

parser = argparse.ArgumentParser(
    description="Extract a SVG from a lottie/tgs file"
)
parser.add_argument(
    "infile",
    help="Input file"
)
parser.add_argument(
    "outfile",
    default="-",
    help="Input file"
)
parser.add_argument(
    "--frame",
    "-f",
    type=float,
    default=0,
    help="(Frame) time at which to get the snapshot"
)
parser.add_argument(
    "--percent",
    "-p",
    type=float,
    default=None,
    help="(Frame) time at which to get the snapshot, as a percentage (eg 50 => middle frame)"
)

if __name__ == "__main__":
    ns = parser.parse_args()

    an = parse_tgs(ns.infile)

    time = ns.frame
    if ns.percent is not None:
        time = an.out_point * ns.percent / 100

    outfile = ns.outfile
    if outfile == "-":
        outfile = sys.stdout
    export_svg(an, outfile, time)



import math
from .base import TgsObject, TgsProp, TgsEnum, NVector
from .properties import Value, MultiDimensional, GradientColors, ShapeProperty, Bezier
from .helpers import Transform


class BoundingBox:
    """!
    Shape bounding box
    """
    def __init__(self, x1=None, y1=None, x2=None, y2=None):
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2

    def include(self, x, y):
        """!
        Expands the box to include the point at x, y
        """
        if x is not None:
            if self.x1 is None or self.x1 > x:
                self.x1 = x
            if self.x2 is None or self.x2 < x:
                self.x2 = x
        if y is not None:
            if self.y1 is None or self.y1 > y:
                self.y1 = y
            if self.y2 is None or self.y2 < y:
                self.y2 = y

    def expand(self, other):
        """!
        Expands the bounding box to include another bounding box
        """
        self.include(other.x1, other.y1)
        self.include(other.x2, other.y2)

    def center(self):
        """!
        Center point of the bounding box
        """
        return NVector((self.x1 + self.x2) / 2, (self.y1 + self.y2) / 2)

    def isnull(self):
        """!
        Whether the box is default-initialized
        """
        return self.x1 is None or self.y2 is None

    def __repr__(self):
        return "<BoundingBox [%s, %s] - [%s, %s]>" % (self.x1, self.y1, self.x2, self.y2)

    @property
    def width(self):
        if self.isnull():
            return 0
        return self.x2 - self.x1

    @property
    def height(self):
        if self.isnull():
            return 0
        return self.y2 - self.y1

    def size(self):
        return NVector(self.width, self.height)


## \ingroup Lottie
class ShapeElement(TgsObject):
    """!
    Base class for all elements of ShapeLayer and Group
    """
    _props = [
        #TgsProp("match_name", "mn", str, False),
        TgsProp("hidden", "hd", bool, False),
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", str, False),
        TgsProp("property_index", "ix", int, False),
    ]
    ## %Shape type.
    type = None
    _shape_classses = None

    def __init__(self):
        # After Effect's Match Name. Used for expressions.
        #self.match_name = ""

        ## After Effect's Name. Used for expressions.
        self.name = None
        ## Property index
        self.property_index = None
        ## Hide element
        self.hidden = None

    def bounding_box(self, time=0):
        """!
        Bounding box of the shape element at the given time
        """
        return BoundingBox()

    @classmethod
    def _load_get_class(cls, lottiedict):
        if not ShapeElement._shape_classses:
            ShapeElement._shape_classses = {}
            ShapeElement._load_sub(ShapeElement._shape_classses)
        return ShapeElement._shape_classses[lottiedict["ty"]]

    @classmethod
    def _load_sub(cls, dict):
        for sc in cls.__subclasses__():
            if sc.type:
                dict[sc.type] = sc
            sc._load_sub(dict)


## \ingroup Lottie
class Shape(ShapeElement):
    """!
    Drawable shape
    """
    _props = [
        TgsProp("direction", "d", float, False),
    ]

    def __init__(self):
        ShapeElement.__init__(self)
        ## After Effect's Direction. Direction how the shape is drawn. Used for trim path for example.
        self.direction = 0

    def to_bezier(self):
        """!
        Returns a Path corresponding to this Shape
        """
        raise NotImplementedError()


## \ingroup Lottie
class Rect(Shape):
    """!
    A simple rectangle shape
    """
    _props = [
        TgsProp("position", "p", MultiDimensional, False),
        TgsProp("size", "s", MultiDimensional, False),
        TgsProp("rounded", "r", Value, False),
    ]
    ## %Shape type.
    type = "rc"

    def __init__(self, pos=None, size=None, rounded=0):
        Shape.__init__(self)
        ## Rect's position
        self.position = MultiDimensional(pos or NVector(0, 0))
        ## Rect's size
        self.size = MultiDimensional(size or NVector(0, 0))
        ## Rect's rounded corners
        self.rounded = Value(rounded)

    def bounding_box(self, time=0):
        pos = self.position.get_value(time)
        sz = self.size.get_value(time)

        return BoundingBox(
            pos[0] - sz[0]/2,
            pos[1] - sz[1]/2,
            pos[0] + sz[0]/2,
            pos[1] + sz[1]/2,
        )

    def to_bezier(self):
        """!
        Returns a Shape corresponding to this rect
        """
        shape = Path()
        kft = set()
        if self.position.animated:
            kft |= set(kf.time for kf in self.position.keyframes)
        if self.size.animated:
            kft |= set(kf.time for kf in self.size.keyframes)
        if self.rounded.animated:
            kft |= set(kf.time for kf in self.rounded.keyframes)
        if not kft:
            shape.shape.value = self._bezier_t(0)
        else:
            for time in sorted(kft):
                shape.shape.add_keyframe(time, self._bezier_t(time))
        return shape

    def _bezier_t(self, time):
        bezier = Bezier()
        bb = self.bounding_box(time)
        rounded = self.rounded.get_value(time)
        tl = NVector(bb.x1, bb.y1)
        tr = NVector(bb.x2, bb.y1)
        br = NVector(bb.x2, bb.y2)
        bl = NVector(bb.x1, bb.y2)

        if not self.rounded.animated and rounded == 0:
            bezier.add_point(tl)
            bezier.add_point(tr)
            bezier.add_point(br)
            bezier.add_point(bl)
        else:
            hh = NVector(rounded/2, 0)
            vh = NVector(0, rounded/2)
            hd = NVector(rounded, 0)
            vd = NVector(0, rounded)
            bezier.add_point(tl+vd, outp=-vh)
            bezier.add_point(tl+hd, -hh)
            bezier.add_point(tr-hd, outp=hh)
            bezier.add_point(tr+vd, -vh)
            bezier.add_point(br-vd, outp=vh)
            bezier.add_point(br-hd, hh)
            bezier.add_point(bl+hd, outp=-hh)
            bezier.add_point(bl-vd, vh)

        bezier.close()
        return bezier


## \ingroup Lottie
class StarType(TgsEnum):
    Star = 1
    Polygon = 2


## \ingroup Lottie
class Star(Shape):
    """!
    Star shape
    """
    _props = [
        TgsProp("position", "p", MultiDimensional, False),
        TgsProp("inner_radius", "ir", Value, False),
        TgsProp("inner_roundness", "is", Value, False),
        TgsProp("outer_radius", "or", Value, False),
        TgsProp("outer_roundness", "os", Value, False),
        TgsProp("rotation", "r", Value, False),
        TgsProp("points", "pt", Value, False),
        TgsProp("star_type", "sy", StarType, False),
    ]
    ## %Shape type.
    type = "sr"

    def __init__(self):
        Shape.__init__(self)
        ## Star's position
        self.position = MultiDimensional(NVector(0, 0))
        ## Star's inner radius. (Star only)
        self.inner_radius = Value()
        ## Star's inner roundness. (Star only)
        self.inner_roundness = Value()
        ## Star's outer radius.
        self.outer_radius = Value()
        ## Star's outer roundness.
        self.outer_roundness = Value()
        ## Star's rotation.
        self.rotation = Value()
        ## Star's number of points.
        self.points = Value(5)
        ## Star's type. Polygon or Star.
        self.star_type = StarType.Star

    def bounding_box(self, time=0):
        pos = self.position.get_value(time)
        r = self.outer_radius.get_value(time)

        return BoundingBox(
            pos[0] - r,
            pos[1] - r,
            pos[0] + r,
            pos[1] + r,
        )

    def to_bezier(self):
        """!
        Returns a Shape corresponding to this star
        """
        shape = Path()
        kft = set()
        if self.position.animated:
            kft |= set(kf.time for kf in self.position.keyframes)
        if self.inner_radius.animated:
            kft |= set(kf.time for kf in self.inner_radius.keyframes)
        if self.inner_roundness.animated:
            kft |= set(kf.time for kf in self.inner_roundness.keyframes)
        if self.points.animated:
            kft |= set(kf.time for kf in self.points.keyframes)
        if self.rotation.animated:
            kft |= set(kf.time for kf in self.rotation.keyframes)
        # TODO inner_roundness / outer_roundness
        if not kft:
            shape.shape.value = self._bezier_t(0)
        else:
            for time in sorted(kft):
                shape.shape.add_keyframe(time, self._bezier_t(time))
        return shape

    def _bezier_t(self, time):
        bezier = Bezier()
        pos = self.position.get_value(time)
        r1 = self.inner_radius.get_value(time)
        r2 = self.outer_radius.get_value(time)
        rot = (self.rotation.get_value(time)) * math.pi / 180 + math.pi
        p = self.points.get_value(time)
        halfd = math.pi / p

        for i in range(int(p)):
            main_angle = rot + i * halfd * 2
            dx = r2 * math.sin(main_angle)
            dy = r2 * math.cos(main_angle)
            bezier.add_point(NVector(pos.x + dx, pos.y + dy))

            if self.star_type == StarType.Star:
                dx = r1 * math.sin(main_angle+halfd)
                dy = r1 * math.cos(main_angle+halfd)
                bezier.add_point(NVector(pos.x + dx, pos.y + dy))

        bezier.close()
        return bezier


## \ingroup Lottie
class Ellipse(Shape):
    """!
    Ellipse shape
    """
    _props = [
        TgsProp("position", "p", MultiDimensional, False),
        TgsProp("size", "s", MultiDimensional, False),
    ]
    ## %Shape type.
    type = "el"

    def __init__(self):
        Shape.__init__(self)
        ## Ellipse's position
        self.position = MultiDimensional(NVector(0, 0))
        ## Ellipse's size
        self.size = MultiDimensional(NVector(0, 0))

    def bounding_box(self, time=0):
        pos = self.position.get_value(time)
        sz = self.size.get_value(time)

        return BoundingBox(
            pos[0] - sz[0]/2,
            pos[1] - sz[1]/2,
            pos[0] + sz[0]/2,
            pos[1] + sz[1]/2,
        )

    def to_bezier(self):
        """!
        Returns a Shape corresponding to this ellipse
        """
        shape = Path()
        kft = set()
        if self.position.animated:
            kft |= set(kf.time for kf in self.position.keyframes)
        if self.size.animated:
            kft |= set(kf.time for kf in self.size.keyframes)
        if not kft:
            shape.shape.value = self._bezier_t(0)
        else:
            for time in sorted(kft):
                shape.shape.add_keyframe(time, self._bezier_t(time))
        return shape

    def _bezier_t(self, time):
        from ..utils.ellipse import Ellipse as EllipseConverter

        bezier = Bezier()
        position = self.position.get_value(time)
        radii = self.size.get_value(time) / 2

        el = EllipseConverter(position, radii, 0)
        points = el.to_bezier(0, math.pi*2)
        for point in points[1:]:
            bezier.add_point(point.vertex, point.in_tangent, point.out_tangent)

        bezier.close()
        return bezier


## \ingroup Lottie
class Path(Shape):
    """!
    Animatable Bezier curve
    """
    _props = [
        TgsProp("shape", "ks", ShapeProperty, False),
    ]
    ## %Shape type.
    type = "sh"

    def __init__(self, bezier=None):
        Shape.__init__(self)
        ## Shape's vertices
        self.shape = ShapeProperty(bezier or Bezier())

    def bounding_box(self, time=0):
        pos = self.shape.get_value(time)

        bb = BoundingBox()
        for v in pos.vertices:
            bb.include(*v)

        return bb

    def to_bezier(self):
        return self


## \ingroup Lottie
class Group(ShapeElement):
    """!
    ShapeElement that can contain other shapes
    @note Shapes inside the same group will create "holes" in other shapes
    """
    _props = [
        TgsProp("number_of_properties", "np", float, False),
        TgsProp("shapes", "it", ShapeElement, True),
    ]
    ## %Shape type.
    type = "gr"

    def __init__(self):
        ShapeElement.__init__(self)
        ## Group number of properties. Used for expressions.
        self.number_of_properties = None
        ## Group list of items
        self.shapes = [TransformShape()]

    def add_shape(self, shape):
        self.shapes.insert(-1, shape)
        return shape

    def insert_shape(self, index, shape):
        self.shapes.insert(index, shape)
        return shape

    @property
    def transform(self):
        return self.shapes[-1]

    def bounding_box(self, time=0):
        bb = BoundingBox()
        for v in self.shapes:
            bb.expand(v.bounding_box(time))
        s = self.transform.scale.get_value(time) / 100
        a = self.transform.anchor_point.get_value(time)
        p = self.transform.position.get_value(time) - a
        r = self.transform.rotation.get_value(time) * math.pi / 180
        if not bb.isnull():
            bb.x1 = bb.x1 * s.x + p.x
            bb.y1 = bb.y1 * s.y + p.y
            bb.x2 = bb.x2 * s.x + p.x
            bb.y2 = bb.y2 * s.y + p.y
            if r:
                bbc = bb.center()
                bbs = bb.size() / 2
                relc = bbc - a
                r += relc.polar_angle
                bbc = a + NVector(math.cos(r), math.sin(r)) * relc.length
                bb = BoundingBox(bbc.x - bbs.x, bbc.y - bbs.y, bbc.x + bbs.x, bbc.y + bbs.y)
        return bb

    def find_all(self, type=ShapeElement, predicate=None, recursive=True):
        """!
        Returns all the child shapes matching a predicate
        """
        results = []
        for e in self.shapes:
            include = isinstance(e, type)
            if predicate and include:
                include = predicate(e)
            if include:
                results.append(e)
            if recursive and isinstance(e, Group):
                results += e.find_all(type, predicate, recursive)
        return results


## \ingroup Lottie
class Fill(ShapeElement):
    """!
    Solid fill color
    """
    _props = [
        TgsProp("opacity", "o", Value, False),
        TgsProp("color", "c", MultiDimensional, False),
    ]
    ## %Shape type.
    type = "fl"

    def __init__(self, color=None):
        ShapeElement.__init__(self)
        ## Fill Opacity
        self.opacity = Value(100)
        ## Fill Color
        self.color = MultiDimensional(color or NVector(1, 1, 1))


## \ingroup Lottie
class GradientType(TgsEnum):
    Linear = 1
    Radial = 2


## \ingroup Lottie
class Gradient(TgsObject):
    _props = [
        TgsProp("start_point", "s", MultiDimensional, False),
        TgsProp("end_point", "e", MultiDimensional, False),
        TgsProp("gradient_type", "t", GradientType, False),
        TgsProp("highlight_length", "h", Value, False),
        TgsProp("highlight_angle", "a", Value, False),
        TgsProp("colors", "g", GradientColors, False),
    ]

    def __init__(self, colors=[]):
        ## Fill Opacity
        self.opacity = Value(100)
        ## Gradient Start Point
        self.start_point = MultiDimensional(NVector(0, 0))
        ## Gradient End Point
        self.end_point = MultiDimensional(NVector(0, 0))
        ## Gradient Type
        self.gradient_type = GradientType.Linear
        ## Gradient Highlight Length. Only if type is Radial
        self.highlight_length = Value()
        ## Highlight Angle. Only if type is Radial
        self.highlight_angle = Value()
        ## Gradient Colors
        self.colors = GradientColors(colors)


## \ingroup Lottie
class GradientFill(ShapeElement, Gradient):
    """!
    Gradient fill
    """
    _props = [
        TgsProp("opacity", "o", Value, False),
    ]
    ## %Shape type.
    type = "gf"

    def __init__(self, colors=[]):
        ShapeElement.__init__(self)
        Gradient.__init__(self, colors)
        ## Fill Opacity
        self.opacity = Value(100)


## \ingroup Lottie
class LineJoin(TgsEnum):
    Miter = 1
    Round = 2
    Bevel = 3


## \ingroup Lottie
class LineCap(TgsEnum):
    Butt = 1
    Round = 2
    Square = 3


## \ingroup Lottie
class BaseStroke(TgsObject):
    _props = [
        TgsProp("line_cap", "lc", LineCap, False),
        TgsProp("line_join", "lj", LineJoin, False),
        TgsProp("miter_limit", "ml", float, False),
        TgsProp("opacity", "o", Value, False),
        TgsProp("width", "w", Value, False),
    ]

    def __init__(self, width=1):
        ## Stroke Line Cap
        self.line_cap = LineCap.Round
        ## Stroke Line Join
        self.line_join = LineJoin.Round
        ## Stroke Miter Limit. Only if Line Join is set to Miter.
        self.miter_limit = 0
        ## Stroke Opacity
        self.opacity = Value(100)
        ## Stroke Width
        self.width = Value(width)


## \ingroup Lottie
class Stroke(ShapeElement, BaseStroke):
    """!
    Solid stroke
    """
    _props = [
        TgsProp("color", "c", MultiDimensional, False),
    ]
    ## %Shape type.
    type = "st"

    def __init__(self, color=None, width=1):
        ShapeElement.__init__(self)
        BaseStroke.__init__(self, width)
        ## Stroke Color
        self.color = MultiDimensional(color or NVector(0, 0, 0))


## \ingroup Lottie
class GradientStroke(ShapeElement, BaseStroke, Gradient):
    """!
    Gradient stroke
    """
    ## %Shape type.
    type = "gs"

    def __init__(self, stroke_width=1):
        ShapeElement.__init__(self)
        BaseStroke.__init__(self, stroke_width)
        Gradient.__init__(self)

    def bounding_box(self, time=0):
        return BoundingBox()


## \ingroup Lottie
class TransformShape(ShapeElement, Transform):
    """!
    Group transform
    """
    ## %Shape type.
    type = "tr"

    def __init__(self):
        ShapeElement.__init__(self)
        Transform.__init__(self)
        self.anchor_point = MultiDimensional(NVector(0, 0))


## \ingroup Lottie
## \todo Implement SVG/SIF Export
class Trim(ShapeElement):
    """
    Trims shapes into a segment
    """
    _props = [
        TgsProp("start", "s", Value, False),
        TgsProp("end", "e", Value, False),
        TgsProp("angle", "o", Value, False),
    ]
    ## %Shape type.
    type = "tm"

    def __init__(self):
        ShapeElement.__init__(self)
        ## Start of the segment, as a percentage
        self.start = Value(0)
        ## End of the segment, as a percentage
        self.end = Value(100)
        ## Angle where to start
        self.angle = Value(0)


## \ingroup Lottie
class Composite(TgsEnum):
    Above = 1
    Below = 2


## \ingroup Lottie
class RepeaterTransform(Transform):
    _props = [
        TgsProp("start_opacity", "so", Value, False),
        TgsProp("end_opacity", "eo", Value, False),
    ]

    def __init__(self):
        Transform.__init__(self)
        self.start_opacity = Value(100)
        self.end_opacity = Value(100)


## \ingroup Lottie
class Modifier(ShapeElement):
    pass


## \ingroup Lottie
class Repeater(Modifier):
    """
    Duplicates previous shapes in a group
    """
    _props = [
        TgsProp("copies", "c", Value, False),
        TgsProp("offset", "o", Value, False),
        TgsProp("composite", "m", Composite, False),
        TgsProp("transform", "tr", RepeaterTransform, False),
    ]
    ## %Shape type.
    type = "rp"

    def __init__(self, copies=1):
        Modifier.__init__(self)
        ## Number of Copies
        self.copies = Value(copies)
        ## Offset of Copies
        self.offset = Value()
        ## Composite of copies
        self.composite = Composite.Above
        ## Transform values for each repeater copy
        self.transform = RepeaterTransform()


## \ingroup Lottie
## \todo Implement SIF Export
class RoundedCorners(Modifier):
    """
    Rounds corners of other shapes
    """
    _props = [
        TgsProp("radius", "r", Value, False),
    ]
    ## %Shape type.
    type = "rd"

    def __init__(self):
        Modifier.__init__(self)
        ## Rounded Corner Radius
        self.radius = Value()


## \ingroup Lottie
## \ingroup LottieCheck
## @note marked as unsupported by lottie
class Merge(ShapeElement):
    _props = [
        TgsProp("merge_mode", "mm", float, False),
    ]
    ## %Shape type.
    type = "mm"

    def __init__(self):
        ShapeElement.__init__(self)
        ## Merge Mode
        self.merge_mode = 1



#!/usr/bin/env python3

import sys
import os
import argparse
sys.path.append(os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    "lib"
))
from tgs.exporters import export_tgs, export_lottie
from tgs.parsers.svg.importer import parse_color

parser = argparse.ArgumentParser(
    description="Vectorizes raster images into a TGS file"
)

parser.add_argument(
    "infile",
    nargs="+",
    help="Input frames"
)
parser.add_argument(
    "--output",
    "-o",
    default="-",
    help="Output file"
)
parser.add_argument(
    "--format",
    "-f",
    default="tgs",
    choices=["tgs", "lottie"],
    help="Output format"
)
parser.add_argument(
    "--framerate",
    default=60,
    type=int,
    help="Output file framerate"
)
parser.add_argument(
    "--delay", "-d",
    default=1,
    type=int,
    help="Delay between source images"
)
parser.add_argument(
    "--colors", "-c",
    default=1,
    type=int,
    help="Number of colors to quantize"
)
parser.add_argument(
    "--palette",
    "-p",
    type=parse_color,
    default=[],
    nargs="+",
    help="Custom palette"
)
parser.add_argument(
    "--color-mode",
    default="nearest",
    choices=["nearest", "exact"],
    help="How to quantize colors." +
         "nearest will map each color to the most similar in the palette." +
         " exact will only match exact colors"
)
parser.add_argument(
    "--mode",
    "-m",
    default="bezier",
    choices=["bezier", "pixel"],
    help="Vectorization mode"
)

if __name__ == "__main__":
    ns = parser.parse_args()
    if ns.mode == "bezier":
        from tgs.parsers.raster import QuanzationMode, raster_to_animation

        cm = QuanzationMode.Nearest if ns.color_mode == "nearest" else QuanzationMode.Exact

        animation = raster_to_animation(
            ns.infile, ns.colors, ns.delay,
            framerate=ns.framerate,
            palette=ns.palette,
            mode=cm
        )
    else:
        from tgs.parsers.pixel import pixel_to_animation
        animation = pixel_to_animation(
            ns.infile, ns.delay, ns.framerate
        )

    binary = ns.format == "tgs"
    if ns.output == "-":
        outfile = sys.stdout
        if binary:
            outfile = outfile.buffer
    else:
        outfile = open(ns.output, "w" + ("b" if binary else ""))

    if ns.format == "lottie":
        export_lottie(animation, outfile, True)
    else:
        export_tgs(animation, outfile)



#!/usr/bin/env python3

import re
import sys
import os
import argparse
sys.path.append(os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    "lib"
))
from tgs.exporters import exporters
from tgs.exporters.base import ExtraOption, _add_options
from tgs import parsers
from tgs.parsers.svg.importer import parse_color
from tgs.utils.stripper import float_strip, heavy_strip
try:
    import tgs.parsers.pixel
    pixel = True
except ImportError:
    pixel = False


class Importer:
    def __init__(self, name, extensions, callback, extra_options=[], slug=None):
        self.name = name
        self.extensions = extensions
        self.callback = callback
        self.extra_options = extra_options
        self.slug = slug if slug is not None else extensions[0]

    def load(self, stream, extra_options):
        return self.callback(stream, **extra_options)


importers = [
    Importer("SVG", ["svg"], parsers.svg.parse_svg_file, [
        ExtraOption("layer_frames", type=int, default=0,
            help="If greater than 0, treats every layer in the SVG as a different animation frame, "
            "greater values increase the time each frames lasts for."),
        ExtraOption("n_frames", type=int, default=60),
        ExtraOption("framerate", type=int, default=60),
    ]),
    Importer("Lottie JSON / Telegram Sticker", ["json", "tgs"], parsers.tgs.parse_tgs, [], "lottie"),
]
if pixel:
    try:
        import tgs.parsers.raster
        raster = True
    except ImportError:
        raster = False

    def bitmap_to_animation(filenames, n_colors, palette, mode, frame_delay=1, framerate=60):
        if raster and mode == "bezier":
            return tgs.parsers.raster.raster_to_animation(
                filenames, n_colors, frame_delay,
                framerate=framerate,
                palette=palette
            )
        elif mode == "polygon":
            return tgs.parsers.pixel.pixel_to_animation_paths(filenames, frame_delay, framerate)
        else:
            return tgs.parsers.pixel.pixel_to_animation(filenames, frame_delay, framerate)

    mode_option = ExtraOption(
        "mode",
        default="pixel",
        choices=["bezier", "pixel", "polygon"],
        help="Vectorization mode"
    )

    importers.append(
        Importer("Raster image", ["bmp", "png", "gif"], bitmap_to_animation, [
            ExtraOption("n_colors", type=int, default=1, help="Number of colors to quantize"),
            ExtraOption("palette", type=parse_color, default=[], nargs="+", help="Custom palette"),
            mode_option,
            ExtraOption("frame_delay", type=int, default=4, help="Number of frames to skip between images"),
            ExtraOption("framerate", type=int, default=60, help="Frames per second"),
            # TODO QuanzationMode for raster
        ])
    )


desc = """Converts between multiple formats

Supported formats:

- Input:
%s

- Output:
%s
""" % (
    "\n".join("%s- %s" % (" "*2, e.name) for e in importers),
    "\n".join("%s- %s" % (" "*2, e.name) for e in exporters),
)

parser = argparse.ArgumentParser(
    description=desc,
    formatter_class=argparse.RawDescriptionHelpFormatter,
)

group = parser.add_argument_group("Generic input options")
group.add_argument(
    "infile",
    default="-",
    nargs="?",
    help="Input file"
)
group.add_argument(
    "--input-format", "-if",
    default=None,
    choices=[importer.slug for importer in importers],
    help="Explicit output format (if missing implied by the input filename)",
)


group = exporters.set_options(parser)

group.add_argument(
    "outfile",
    default="-",
    nargs="?",
    help="Output file"
)
group.add_argument(
    "--output-format", "-of",
    default=None,
    choices=[exporter.slug for exporter in exporters],
    help="Explicit output format (if missing implied by the output filename)",
)
group.add_argument(
    "--sanitize",
    default=False,
    action="store_true",
    help="Ensure the animation is 512x512 and 30 or 60 fps",
)
group.add_argument(
    "--optimize",
    "-O",
    default=1,
    type=int,
    choices=[0, 1, 2],
    help="Optimize the animation parameter: 0 no optimization, 1 truncate floats, 2 truncate floats and names",
)


for importer in importers:
    _add_options(parser, "import", importer)

if __name__ == "__main__":
    ns = parser.parse_args()

    infile = ns.infile
    importer = None
    if infile == "-":
        infile = sys.stdin
    else:
        suf = os.path.splitext(infile)[1][1:]
        for p in importers:
            if suf in p.extensions:
                importer = p
                break
    if ns.input_format:
        importer = None
        for p in importers:
            if p.slug == ns.input_format:
                importer = p
                break
    if not importer:
        sys.stderr.write("Unknown importer\n")
        sys.exit(1)

    outfile = ns.outfile
    exporter = None
    if outfile == "-":
        outfile = sys.stdout
    else:
        exporter = exporters.get_from_filename(outfile)
    if ns.output_format:
        exporter = exporters.get(ns.output_format)

    if not exporter:
        sys.stderr.write("Unknown exporter\n")
        sys.exit(1)

    i_options = {}
    for opt in importer.extra_options:
        i_options[opt.name] = getattr(ns, opt.nsvar(importer.slug))

    o_options = exporter.argparse_options(ns)

    an = importer.load(infile, i_options)
    if ns.optimize == 1:
        float_strip(an)
    elif ns.optimize >= 2:
        heavy_strip(an)
    if ns.sanitize:
        an.tgs_sanitize()
    exporter.export(an, outfile, o_options)



import cairosvg
import io

from .base import exporter
from .svg import export_svg


def _export_cairo(func, animation, fp, frame, dpi):
    intermediate = io.StringIO()
    export_svg(animation, intermediate, frame)
    intermediate.seek(0)
    func(file_obj=intermediate, write_to=fp, dpi=dpi)


@exporter("PNG", ["png"], [], {"frame"})
def export_png(animation, fp, frame=0, dpi=96):
    _export_cairo(cairosvg.svg2png, animation, fp, frame, dpi)


@exporter("PDF", ["pdf"], [], {"frame"})
def export_pdf(animation, fp, frame=0, dpi=96):
    _export_cairo(cairosvg.svg2pdf, animation, fp, frame, dpi)


@exporter("PostScript", ["ps"], [], {"frame"})
def export_ps(animation, fp, frame=0, dpi=96):
    _export_cairo(cairosvg.svg2ps, animation, fp, frame, dpi)




from .base import exporter
from ..parsers.sif.builder import to_sif


@exporter("Synfig", ["sif"], [], {"pretty"})
def export_sif(animation, fp, pretty=True):
    dom = to_sif(animation)
    if isinstance(fp, str):
        fp = open(fp, "w")
    dom.writexml(fp, "", "  " if pretty else "", "\n" if pretty else "")



from .base import TgsObject, TgsProp, TgsEnum
from .properties import Value, MultiDimensional
from ..nvector import NVector


## \ingroup Lottie
## \ingroup LottieCheck
class MaskedPath(TgsObject):
    _props = [
        TgsProp("mask", "m", float),
        TgsProp("f", "f", Value),
        TgsProp("l", "l", Value),
        TgsProp("r", "r", float),
    ]

    def __init__(self):
        ## Type?
        self.mask = None
        ## First?
        self.f = None
        ## Last?
        self.l = None
        ## ??
        self.r = None


## \ingroup Lottie
## \ingroup LottieCheck
class TextAnimatorDataProperty(TgsObject):
    _props = [
        TgsProp("r", "r", Value),
        TgsProp("rx", "rx", Value),
        TgsProp("ry", "ry", Value),
        TgsProp("sk", "sk", Value),
        TgsProp("sa", "sa", Value),
        TgsProp("s", "s", MultiDimensional),
        TgsProp("a", "a", MultiDimensional),
        TgsProp("o", "o", Value),
        TgsProp("p", "p", MultiDimensional),
        TgsProp("sw", "sw", Value),
        TgsProp("sc", "sc", MultiDimensional),
        TgsProp("fc", "fc", MultiDimensional),
        TgsProp("fh", "fh", Value),
        TgsProp("fs", "fs", Value),
        TgsProp("fb", "fb", Value),
        TgsProp("t", "t", Value),
    ]

    def __init__(self):
        ## Angle?
        self.r = Value()
        self.rx = Value()
        ## Angle?
        self.ry = Value()
        ## Angle?
        self.sk = Value()
        ## Angle?
        self.sa = Value()
        ## Scale? 0-100?
        self.s = MultiDimensional()
        self.a = MultiDimensional()
        ## Opacity? 0-100?
        self.o = Value()
        self.p = MultiDimensional()
        self.sw = Value()
        self.sc = MultiDimensional()
        self.fc = MultiDimensional()
        self.fh = Value()
        ## 0-100?
        self.fs = Value()
        ## 0-100?
        self.fb = Value()
        self.t = Value()


## \ingroup Lottie
## \ingroup LottieCheck
class TextMoreOptions(TgsObject):
    _props = [
        TgsProp("alignment", "a", MultiDimensional),
        TgsProp("g", "g", float),
    ]

    def __init__(self):
        self.alignment = MultiDimensional(NVector(0, 0))
        self.g = None


## \ingroup Lottie
class TextJustify(TgsEnum):
    Left = 0
    Right = 1
    Center = 2


## \ingroup Lottie
class TextDocument(TgsObject):
    """!
    @see http://docs.aenhancers.com/other/textdocument/
    """
    _props = [
        TgsProp("font_family", "f", str),
        TgsProp("color", "fc", NVector),
        TgsProp("font_size", "s", float),
        TgsProp("line_height", "lh", float),
        TgsProp("wrap_size", "sz", NVector),
        TgsProp("text", "t", str),
        TgsProp("justify", "j", TextJustify),
        # ls?
    ]

    def __init__(self, text="", font_size=10, color=None, font_family=""):
        self.font_family = font_family
        ## Text color
        self.color = color or NVector(0, 0, 0)
        ## Line height when wrapping
        self.line_height = None
        ## Text alignment
        self.justify = TextJustify.Left
        ## Size of the box containing the text
        self.wrap_size = None
        ## Text
        self.text = text
        ## Font Size
        self.font_size = font_size


## \ingroup Lottie
class TextDataKeyframe(TgsObject):
    _props = [
        TgsProp("start", "s", TextDocument),
        TgsProp("time", "t", float),
    ]

    def __init__(self, time=0, start=None):
        ## Start value of keyframe segment.
        self.start = start
        ## Start time of keyframe segment.
        self.time = time


## \ingroup Lottie
class TextData(TgsObject):
    _props = [
        TgsProp("keyframes", "k", TextDataKeyframe, True),
    ]

    def __init__(self):
        self.keyframes = []


## \ingroup Lottie
class TextAnimatorData(TgsObject):
    _props = [
        TgsProp("properties", "a", TextAnimatorDataProperty, True),
        TgsProp("data", "d", TextData, False),
        TgsProp("more_options", "m", TextMoreOptions, False),
        TgsProp("masked_path", "p", MaskedPath),
    ]

    def __init__(self):
        self.properties = []
        self.data = TextData()
        self.more_options = TextMoreOptions()
        self.masked_path = MaskedPath()

    def add_keyframe(self, time, item):
        self.data.keyframes.append(TextDataKeyframe(time, item))


## \ingroup Lottie
class Font(TgsObject):
    _props = [
        TgsProp("ascent", "ascent", float),
        TgsProp("font_family", "fFamily", str),
        TgsProp("name", "fName", str),
        TgsProp("font_style", "fStyle", str),
    ]

    def __init__(self, font_family="sans", font_style="Regular", name=None):
        self.ascent = None
        self.font_family = font_family
        self.font_style = font_style
        self.name = name or "%s-%s" % (font_family, font_style)


## \ingroup Lottie
class FontList(TgsObject):
    _props = [
        TgsProp("list", "list", Font, True),
    ]

    def __init__(self):
        self.list = []

    def append(self, font):
        self.list.append(font)



import io
from PIL import Image

from .cairo import export_png
from .base import exporter, ExtraOption


def _png_gif_prepare(image):
    if image.mode not in ["RGBA", "RGBa"]:
        image = image.convert("RGBA")
    alpha = image.getchannel("A")
    image = image.convert('P', palette=Image.ADAPTIVE, colors=255)
    mask = Image.eval(alpha, lambda a: 255 if a <= 128 else 0)
    image.paste(255, mask)
    return image


@exporter("GIF", ["gif"], [
    ExtraOption("skip_frames", type=int, default=5, help="Only renderer 1 out of these many frames"),
])
def export_gif(animation, fp, dpi=96, skip_frames=5):
    """
    Gif export

    Note that it's a bit slow.
    """
    start = int(animation.in_point)
    end = int(animation.out_point)
    frames = []
    for i in range(start, end+1, skip_frames):
        file = io.BytesIO()
        export_png(animation, file, i, dpi)
        file.seek(0)
        frames.append(_png_gif_prepare(Image.open(file)))

    duration = 1000 / animation.frame_rate
    frames[0].save(
        fp,
        format='GIF',
        append_images=frames[1:],
        save_all=True,
        duration=duration,
        loop=0,
        transparency=255,
        disposal=2,
    )



import os
import subprocess
import fontTools.pens.basePen
import fontTools.ttLib
import enum
import math
from xml.etree import ElementTree
from ..nvector import NVector
from ..objects.bezier import Bezier, BezierPoint
from ..objects.shapes import Path, Group, Fill, Stroke
from ..objects.text import TextJustify
from ..objects.base import TgsProp, CustomObject


class BezierPen(fontTools.pens.basePen.BasePen):
    def __init__(self, glyphSet, offset=NVector(0, 0)):
        super().__init__(glyphSet)
        self.beziers = []
        self.current = Bezier()
        self.offset = offset

    def _point(self, pt):
        return self.offset + NVector(pt[0], -pt[1])

    def _moveTo(self, pt):
        self._endPath()

    def _endPath(self):
        if len(self.current.points):
            self.beziers.append(self.current)
        self.current = Bezier()

    def _closePath(self):
        self.current.close()
        self._endPath()

    def _lineTo(self, pt):
        if len(self.current.points) == 0:
            self.current.points.append(self._point(self._getCurrentPoint()))

        self.current.points.append(self._point(pt))

    def _curveToOne(self, pt1, pt2, pt3):
        if len(self.current.points) == 0:
            cp = self._point(self._getCurrentPoint())
            self.current.points.append(
                BezierPoint(
                    cp,
                    None,
                    self._point(pt1) - cp
                )

            )
        else:
            self.current.points[-1].out_tangent = self._point(pt1) - self.current.points[-1].vertex

        dest = self._point(pt3)
        self.current.points.append(
            BezierPoint(
                dest,
                self._point(pt2) - dest,
                None,
            )
        )


class SystemFont:
    def __init__(self, family):
        self.family = family
        self.files = {}
        self.styles = set()
        self._renderers = {}

    def add_file(self, styles, file):
        self.styles |= set(styles)
        self.files[tuple(sorted(styles))] = file

    def filename(self, styles):
        return self.files[self._key(styles)]

    def _key(self, styles):
        if isinstance(styles, str):
            return (styles,)
        return tuple(sorted(styles))

    def __getitem__(self, styles):
        key = self._key(styles)
        if key in self._renderers:
            return self._renderers[key]
        fr = FontRenderer(self.files[key])
        self._renderers[key] = fr
        return fr

    def __repr__(self):
        return "<SystemFont %s>" % self.family


class FontQuery:
    """!
    \see https://www.freedesktop.org/software/fontconfig/fontconfig-user.html#AEN21
         https://manpages.ubuntu.com/manpages/cosmic/man1/fc-pattern.1.html
    """
    def __init__(self, str=""):
        self._query = {}
        if isinstance(str, FontQuery):
            self._query = str._query.copy()
        elif str:
            chunks = str.split(":")
            family = chunks.pop(0)
            self._query = dict(
                chunk.split("=")
                for chunk in chunks
                if chunk
            )
            self.family(family)

    def family(self, name):
        self._query["family"] = name
        return self

    def weight(self, weight):
        self._query["weight"] = weight
        return self

    def css_weight(self, weight):
        """!
            Weight from CSS weight value.

            Weight is different between CSS and fontconfig
            This creates some interpolations to ensure known values are translated properly
            \see https://www.freedesktop.org/software/fontconfig/fontconfig-user.html#AEN178
                 https://developer.mozilla.org/en-US/docs/Web/CSS/font-weight#Common_weight_name_mapping
        """
        if weight < 200:
            v = max(0, weight - 100) / 100 * 40
        elif weight < 500:
            v = -weight**3 / 200000 + weight**2 * 11/2000 - weight * 17/10 + 200
        elif weight < 700:
            v = -weight**2 * 3/1000 + weight * 41/10 - 1200
        else:
            v = (weight - 700) / 200 * 10 + 200
        return self.weight(int(round(v)))

    def style(self, *styles):
        self._query["style"] = " ".join(styles)
        return self

    def charset(self, *hex_ranges):
        self._query["charset"] = " ".join(hex_ranges)
        return self

    def char(self, char):
        return self.charset("%x" % ord(char))

    def custom(self, property, value):
        self._query[property] = value
        return self

    def clone(self):
        return FontQuery(self)

    def __getitem__(self, key):
        return self._query.get(key, "")

    def __contains__(self, item):
        return item in self._query

    def get(self, key, default=None):
        return self._query.get(key, default)

    def __str__(self):
        return self._query.get("family", "") + ":" + ":".join(
            "%s=%s" % (p, v)
            for p, v in self._query.items()
            if p != "family"
        )

    def __repr__(self):
        return "<FontQuery %r>" % str(self)

    def weight_to_css(self):
        x = int(self["weight"])
        if x < 40:
            v = x / 40 * 100 + 100
        elif x < 100:
            v = x**3/300 - x**2 * 11/15 + x*167/3 - 3200/3
        elif x < 200:
            v = (2050 - 10 * math.sqrt(5) * math.sqrt(1205 - 6 * x)) / 3
        else:
            v = (x - 200) * 200 / 10 + 700
        return int(round(v))


class _SystemFontList:
    def __init__(self):
        self.fonts = None

    def _lazy_load(self):
        if self.fonts is None:
            self.load()

    def load(self):
        self.fonts = {}
        self.load_fc_list()

    def cmd(self, *a):
        p = subprocess.Popen(a, stdout=subprocess.PIPE)
        out, err = p.communicate()
        out = out.decode("utf-8").strip()
        return out, p.returncode

    def load_fc_list(self):
        out, returncode = self.cmd("fc-list", r'--format=%{file}\t%{family[0]}\t%{style[0]}\n')
        if returncode == 0:
            for line in out.splitlines():
                file, family, styles = line.split("\t")
                self._get(family).add_file(styles.split(" "), file)

    def best(self, query):
        """!
        Returns the renderer best matching the name
        """
        out, returncode = self.cmd("fc-match", r"--format=%{family}\t%{style}", str(query))
        if returncode == 0:
            fam, style = out.split("\t")
            return self[fam][style]

    def default(self):
        """!
        Returns the default fornt renderer
        """
        return self.best()

    def _get(self, family):
        self._lazy_load()
        if family in self.fonts:
            return self.fonts[family]
        font = SystemFont(family)
        self.fonts[family] = font
        return font

    def __getitem__(self, key):
        self._lazy_load()
        return self.fonts[key]

    def __iter__(self):
        self._lazy_load()
        return iter(self.fonts.values())

    def keys(self):
        self._lazy_load()
        return self.fonts.keys()

    def __contains__(self, item):
        self._lazy_load()
        return item in self.fonts


## Dictionary of system fonts
fonts = _SystemFontList()


class FontRenderer:
    def __init__(self, filename):
        self.filename = filename
        self.font = fontTools.ttLib.TTFont(filename)
        self.glyphset = self.font.getGlyphSet()
        self.cmap = self.font.getBestCmap() or {}

    def glyph_beziers(self, name, offset=NVector(0, 0)):
        pen = BezierPen(self.glyphset, offset)
        self.glyphset[name].draw(pen)
        return pen.beziers

    def glyph_shapes(self, name, offset=NVector(0, 0)):
        beziers = self.glyph_beziers(name, offset)
        return [
            Path(bez)
            for bez in beziers
        ]

    def glyph_group(self, name):
        group = Group()
        group.name = name
        group.shapes = self.glyph_shapes(name) + group.shapes
        return group

    def render(self, text, size, pos=None, on_missing=None):
        """!
        Renders some text

        @param text         String to render
        @param size         Font size (in pizels)
        @param[in,out] pos  Text position
        @param on_missing   Callable on missing glyphs, called with:
        - Character as string
        - Font size
        - [in, out] Character position
        - Group shape

        @returns a Group shape, augmented with some extra attributes:
        - line_height   Line height

        """
        scale = size / self.font.tables["head"].unitsPerEm
        line_height = self.font.tables["head"].yMax * scale
        group = Group()
        group.name = text
        if pos is None:
            pos = NVector(0, 0)
        #group.transform.scale.value = NVector(100, 100) * scale
        for ch in text:
            if ch == "\n":
                pos.x = 0
                pos.y += line_height
                continue

            chname = self.cmap.get(ord(ch)) or self.font._makeGlyphName(ord(ch))
            if chname in self.glyphset:
                for sh in self.glyph_shapes(chname, pos / scale):
                    sh.shape.value.scale(scale)
                    group.add_shape(sh)
                pos.x += self.glyphset[chname].width * scale
            elif on_missing:
                on_missing(ch, size, pos, group)

        group.line_height = line_height
        return group

    def __repr__(self):
        return "<FontRenderer %r>" % self.filename


class FallbackFontRenderer:
    def __init__(self, query):
        self.query = FontQuery(query)
        self._best = None
        self._bq = None

    @property
    def best(self):
        cq = str(self.query)
        if self._best is None or self._bq != cq:
            self._best = fonts.best(self.query)
            self._bq = cq
        return self._best

    def _on_missing(self, char, size, pos, group):
        font = fonts.best(self.query.clone().char(char))
        group.add_shape(font.render(char, size, pos))

    def render(self, text, size, pos=None):
        return self.best.render(text, size, pos, self._on_missing)

    def __repr__(self):
        return "<FallbackFontRenderer %s>" % self.query


class FontStyle:
    def __init__(self, query, size, justify=TextJustify.Left, position=None):
        self._renderer = FallbackFontRenderer(query)
        self.size = size
        self.justify = justify
        self.position = position.clone() if position else NVector(0, 0)

    @property
    def query(self):
        return self._renderer.query

    @query.setter
    def query(self, value):
        if str(value) != str(self.query):
            self._renderer = FallbackFontRenderer(value)

    @property
    def renderer(self):
        return self.renderer

    def render(self, text, pos=NVector(0, 0)):
        group = self._renderer.render(text, self.size, self.position+pos)
        if self.justify == TextJustify.Center:
            group.transform.position.value.x += group.bounding_box().width / 2
        elif self.justify == TextJustify.Right:
            group.transform.position.value.x += group.bounding_box().width
        return group

    def clone(self):
        return FontStyle(str(self._renderer.query), self.size, self.justify)


def _propfac(a):
    return property(lambda s: s._get(a), lambda s, v: s._set(a, v))


class FontShape(CustomObject):
    _props = [
        TgsProp("query_string", "_query", str),
        TgsProp("size", "_size", float),
        TgsProp("justify", "_justify", TextJustify),
        TgsProp("text", "_text", str),
    ]
    wrapped_tgs = Group

    def __init__(self, text="", query="", size=64, justify=TextJustify.Left):
        CustomObject.__init__(self)
        if isinstance(query, FontStyle):
            self.style = query
        else:
            self.style = FontStyle(query, size, justify)
        self.text = text

    def _get(self, a):
        return getattr(self.style, a)

    def _set(self, a, v):
        return setattr(self.style, a, v)

    query = _propfac("query")
    size = _propfac("size")
    justify = _propfac("justify")

    @property
    def query_string(self):
        return str(self.query)

    @query_string.setter
    def query_string(self, v):
        self.query = v

    def _build_wrapped(self):
        g = self.style.render(self.text)
        self.line_height = g.line_height
        return g

    def bounding_box(self, time=0):
        return self.wrapped.bounding_box(time)



import math
from functools import reduce
from .base import TgsObject, TgsProp, PseudoList, PseudoBool
from . import easing
from ..nvector import NVector
from .bezier import Bezier


class KeyframeBezier:
    NEWTON_ITERATIONS = 4
    NEWTON_MIN_SLOPE = 0.001
    SUBDIVISION_PRECISION = 0.0000001
    SUBDIVISION_MAX_ITERATIONS = 10
    SPLINE_TABLE_SIZE = 11
    SAMPLE_STEP_SIZE = 1.0 / (SPLINE_TABLE_SIZE - 1.0)

    def __init__(self, h1, h2):
        self.h1 = h1
        self.h2 = h2
        self._sample_values = None

    @classmethod
    def from_keyframe(cls, keyframe):
        return cls(keyframe.out_value, keyframe.in_value)

    def bezier(self):
        bez = Bezier()
        bez.add_point(NVector(0, 0), outp=NVector(self.h1.x, self.h1.y))
        bez.add_point(NVector(1, 1), inp=NVector(self.h2.x-1, self.h2.y-1))
        return bez

    def _a(self, c1, c2):
        return 1 - 3 * c2 + 3 * c1

    def _b(self, c1, c2):
        return 3 * c2 - 6 * c1

    def _c(self, c1):
        return 3 * c1

    def _bezier_component(self, t, c1, c2):
        return ((self._a(c1, c2) * t + self._b(c1, c2)) * t + self._c(c1)) * t

    def point_at(self, t):
        return NVector(
            self._bezier_component(t, self.h1.x, self.h2.x),
            self._bezier_component(t, self.h1.y, self.h2.y)
        )

    def _slope_component(self, t, c1, c2):
        return 3 * self._a(c1, c2) * t * t + 2 * self._b(c1, c2) * t + self._c(c1)

    def slope_at(self, t):
        return NVector(
            self._slope_component(t, self.h1.x, self.h2.x),
            self._slope_component(t, self.h1.y, self.h2.y)
        )

    def _binary_subdivide(self, x, interval_start, interval_end):
        current_x = None
        t = None
        i = 0
        for i in range(self.SUBDIVISION_MAX_ITERATIONS):
            if current_x is not None and abs(current_x) < self.SUBDIVISION_PRECISION:
                break
            t = interval_start + (interval_end - interval_start) / 2.0
            current_x = self._bezier_component(t, self.h1.x, self.h2.x) - x
            if current_x > 0.0:
                interval_end = t
            else:
                interval_start = t
        return t

    def _newton_raphson(self, x, t_guess):
        for i in range(self.NEWTON_ITERATIONS):
            slope = self._slope_component(t_guess, self.h1.x, self.h2.x)
            if slope == 0:
                return t_guess
            current_x = self._bezier_component(t_guess, self.h1.x, self.h2.x) - x
            t_guess -= current_x / slope
        return t_guess

    def _get_sample_values(self):
        if self._sample_values is None:
            self._sample_values = [
                self._bezier_component(i * self.SAMPLE_STEP_SIZE, self.h1.x, self.h2.x)
                for i in range(self.SPLINE_TABLE_SIZE)
            ]
        return self._sample_values

    def t_for_x(self, x):
        sample_values = self._get_sample_values()
        interval_start = 0
        current_sample = 1
        last_sample = self.SPLINE_TABLE_SIZE - 1
        while current_sample != last_sample and sample_values[current_sample] <= x:
            interval_start += self.SAMPLE_STEP_SIZE
            current_sample += 1
        current_sample -= 1

        dist = (x - sample_values[current_sample]) / (sample_values[current_sample+1] - sample_values[current_sample])
        t_guess = interval_start + dist * self.SAMPLE_STEP_SIZE
        initial_slope = self._slope_component(t_guess, self.h1.x, self.h2.x)
        if initial_slope >= self.NEWTON_MIN_SLOPE:
            return self._newton_raphson(x, t_guess)
        if initial_slope == 0:
            return t_guess
        return self._binary_subdivide(x, interval_start, interval_start + self.SAMPLE_STEP_SIZE)

    def y_at_x(self, x):
        t = self.t_for_x(x)
        return self._bezier_component(t, self.h1.y, self.h2.y)


## \ingroup Lottie
class Keyframe(TgsObject):
    _props = [
        TgsProp("time", "t", float, False),
        TgsProp("in_value", "i", easing.KeyframeBezierHandle, False),
        TgsProp("out_value", "o", easing.KeyframeBezierHandle, False),
        TgsProp("jump", "h", PseudoBool),
    ]

    def __init__(self, time=0, easing_function=None):
        """!
        \param time             Start time of keyframe segment
        \param easing_function  Callable that performs the easing
        """
        ## Start time of keyframe segment.
        self.time = time
        ## Bezier curve easing in value.
        self.in_value = None
        ## Bezier curve easing out value.
        self.out_value = None
        ## Jump to the end value
        self.jump = None

        if easing_function:
            easing_function(self)

    def bezier(self):
        if self.jump:
            bez = Bezier()
            bez.add_point(NVector(0, 0))
            bez.add_point(NVector(1, 0))
            bez.add_point(NVector(1, 1))
            return bez
        else:
            return KeyframeBezier.from_keyframe(self).bezier()

    def lerp_factor(self, ratio):
        return KeyframeBezier.from_keyframe(self).y_at_x(ratio)


## \ingroup Lottie
class OffsetKeyframe(Keyframe):
    """!
    Keyframe for MultiDimensional values

    @par Bezier easing
    @parblock
    Imagine a quadratic bezier, with starting point at (0, 0) and end point at (1, 1).

    \p out_value and \p in_value are the other two handles for a quadratic bezier,
    expressed as absoulte values in this 0-1 space.

    See also https://cubic-bezier.com/
    @endparblock
    """
    _props = [
        TgsProp("start", "s", NVector, False),
        TgsProp("end", "e", NVector, False),
        TgsProp("in_tan", "ti", float, True),
        TgsProp("out_tan", "to", float, True),
    ]

    def __init__(self, time=0, start=None, end=None, easing_function=None):
        Keyframe.__init__(self, time, easing_function)
        ## Start value of keyframe segment.
        self.start = start
        ## End value of keyframe segment.
        self.end = end
        ## In Spatial Tangent. Only for spatial properties. Array of numbers.
        self.in_tan = None
        ## Out Spatial Tangent. Only for spatial properties. Array of numbers.
        self.out_tan = None

    def interpolated_value(self, ratio, next_start=None):
        end = next_start if self.end is None else self.end
        if end is None:
            return self.start
        if not self.in_value or not self.out_value:
            return self.start
        if ratio == 1:
            return end
        if ratio == 0:
            return self.start
        lerpv = self.lerp_factor(ratio)
        return self.start.lerp(end, lerpv)


class AnimatableMixin:
    keyframe_type = Keyframe

    def __init__(self, value=None):
        ## Non-animated value
        self.value = value
        ## Property index
        self.property_index = None
        ## Whether it's animated
        self.animated = False
        ## Keyframe list
        self.keyframes = None

    def clear_animation(self, value):
        """!
        Sets a fixed value, removing animated keyframes
        """
        self.value = value
        self.animated = False
        self.keyframes = None

    def add_keyframe(self, time, value, interp=easing.Linear()):
        """!
        \param time     The time this keyframe appears in
        \param value    The value the property should have at \p time
        \param interp   The easing callable used to update the tangents of the previous keyframe
        @note Always call add_keyframe with increasing \p time value
        """
        if not self.animated:
            self.value = None
            self.keyframes = []
            self.animated = True
        else:
            if self.keyframes[-1].time == time:
                if value != self.keyframes[-1].start:
                    self.keyframes[-1].start = value
                return
            else:
                self.keyframes[-1].end = value.clone()

        self.keyframes.append(self.keyframe_type(
            time,
            value,
            None,
            interp
        ))

    def get_value(self, time=0):
        """!
        @brief Returns the value of the property at the given frame/time
        \todo honour easing
        """
        if not self.animated:
            return self.value

        if not self.keyframes:
            return None

        val = self.keyframes[0].start
        for i in range(len(self.keyframes)):
            k = self.keyframes[i]
            if time - k.time <= 0:
                if k.start is not None:
                    val = k.start

                kp = self.keyframes[i-1] if i > 0 else None
                if kp:
                    end = kp.end
                    if end is None and i + 1 < len(self.keyframes):
                        end = self.keyframes[i+1].start
                    if end is not None:
                        val = kp.interpolated_value((time - kp.time) / (k.time - kp.time), end)
                break
            if k.end is not None:
                val = k.end
        return val

    def to_dict(self):
        d = super().to_dict()
        if self.animated:
            last = d["k"][-1]
            last.pop("i", None)
            last.pop("o", None)
        return d


## \ingroup Lottie
class MultiDimensional(AnimatableMixin, TgsObject):
    """!
    An animatable property that holds a NVector
    """
    keyframe_type = OffsetKeyframe
    _props = [
        TgsProp("value", "k", NVector, False, lambda l: not l.get("a", None)),
        TgsProp("property_index", "ix", int, False),
        TgsProp("animated", "a", PseudoBool, False),
        TgsProp("keyframes", "k", OffsetKeyframe, True, lambda l: l.get("a", None)),
    ]


## \ingroup Lottie
# \todo use a more convenient representation and convert to the weird array on import/export
class GradientColors(TgsObject):
    """!
    Represents colors and offsets in a gradient
    """
    _props = [
        TgsProp("colors", "k", MultiDimensional),
        TgsProp("count", "p", int),
    ]

    def __init__(self, colors=[]):
        ## Animatable colors, as a vector containing [offset, r, g, b] values as a flat array
        self.colors = MultiDimensional(NVector())
        ## Number of colors
        self.count = 0
        if colors:
            self.set_colors(colors)

    def set_colors(self, colors, keyframe=None):
        flat = self._flatten_colors(colors)
        if self.colors.animated and keyframe is not None:
            if keyframe > 1:
                self.colors.keyframes[keyframe-1].end = flat
            self.colors.keyframes[keyframe].start = flat
        else:
            self.colors.clear_animation(flat)
        self.count = len(colors)

    def _flatten_colors(self, colors):
        return NVector(*reduce(
            lambda a, b: a + b,
            map(
                lambda it: [it[0] / (len(colors)-1)] + it[1].components,
                enumerate(colors),
            )
        ))

    def add_color(self, offset, color, keyframe=None):
        flat = [offset] + color
        if self.colors.animated:
            if keyframe is None:
                for kf in self.colors.keyframes:
                    if kf.start:
                        kf.start += flat
                    if kf.end:
                        kf.end += flat
            else:
                if keyframe > 1:
                    self.colors.keyframes[keyframe-1].end.components += flat
                self.colors.keyframes[keyframe].start.components += flat
        else:
            self.colors.value.components += flat
        self.count += 1

    def add_keyframe(self, time, colors=None, ease=easing.Linear()):
        self.colors.add_keyframe(time, self._flatten_colors(colors) if colors else NVector(), ease)


## \ingroup Lottie
class Value(AnimatableMixin, TgsObject):
    """!
    An animatable property that holds a float
    """
    keyframe_type = OffsetKeyframe
    _props = [
        TgsProp("value", "k", float, False, lambda l: not l.get("a", None)),
        TgsProp("property_index", "ix", int, False),
        TgsProp("animated", "a", PseudoBool, False),
        TgsProp("keyframes", "k", keyframe_type, True, lambda l: l.get("a", None)),
    ]

    def __init__(self, value=0):
        super().__init__(value)

    def add_keyframe(self, time, value, ease=easing.Linear()):
        super().add_keyframe(time, NVector(value), ease)

    def get_value(self, time=0):
        v = super().get_value(time)
        if self.animated and self.keyframes:
            return v[0]
        return v


## \ingroup Lottie
class ShapePropKeyframe(Keyframe):
    """!
    Keyframe holding Bezier objects
    """
    _props = [
        TgsProp("start", "s", Bezier, PseudoList),
        TgsProp("end", "e", Bezier, PseudoList),
    ]

    def __init__(self, time=0, start=None, end=None, easing_function=None):
        Keyframe.__init__(self, time, easing_function)
        ## Start value of keyframe segment.
        self.start = start
        ## End value of keyframe segment.
        self.end = end

    def interpolated_value(self, ratio, next_start=None):
        end = next_start if self.end is None else self.end
        if end is None:
            return self.start
        if not self.in_value or not self.out_value:
            return self.start
        if ratio == 1:
            return end
        if ratio == 0 or len(self.start.vertices) != len(end.vertices):
            return self.start

        lerpv = self.lerp_factor(ratio)
        bez = Bezier()
        bez.closed = self.start.closed
        for i in range(len(self.start.vertices)):
            bez.vertices.append(self.start.vertices[i].lerp(end.vertices[i], lerpv))
            bez.in_tangents.append(self.start.in_tangents[i].lerp(end.in_tangents[i], lerpv))
            bez.out_tangents.append(self.start.out_tangents[i].lerp(end.out_tangents[i], lerpv))
        return bez


## \ingroup Lottie
class ShapeProperty(AnimatableMixin, TgsObject):
    """!
    An animatable property that holds a Bezier
    """
    keyframe_type = ShapePropKeyframe
    _props = [
        TgsProp("value", "k", Bezier, False, lambda l: not l["a"]),
        #TgsProp("expression", "x", str, False),
        TgsProp("property_index", "ix", float, False),
        TgsProp("animated", "a", PseudoBool, False),
        TgsProp("keyframes", "k", keyframe_type, True, lambda l: l["a"]),
    ]

    def __init__(self, bezier=None):
        super().__init__(bezier or Bezier())



import re
import math
from xml.etree import ElementTree

from .handler import SvgHandler, NameMode
from ... import objects
from ...nvector import NVector
from ...utils import restructure
try:
    from ...utils import font
    has_font = True
except ImportError:
    has_font = False


class SvgBuilder(SvgHandler, restructure.AbstractBuilder):
    merge_paths = True
    namestart = (
        r":_A-Za-z\xC0-\xD6\xD8-\xF6\xF8-\u02FF\u0370-\u037D\u037F-\u1FFF" +
        r"\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF" +
        r"\uFDF0-\uFFFD\U00010000-\U000EFFFF"
    )
    namenostart = r"-.0-9\xB7\u0300-\u036F\u203F-\u2040"
    id_re = re.compile("^[%s][%s%s]*$" % (namestart, namenostart, namestart))

    def __init__(self, time=0):
        super().__init__()
        self.svg = ElementTree.Element("svg")
        self.dom = ElementTree.ElementTree(self.svg)
        self.svg.attrib["xmlns"] = self.ns_map["svg"]
        self.ids = set()
        self.idc = 0
        self.name_mode = NameMode.Inkscape
        self.time = time

    def gen_id(self, prefix="id"):
        #TODO should check if id_n already exists
        self.idc += 1
        id = "%s_%s" % (prefix, self.idc)
        self.ids.add(id)
        return id

    def set_id(self, dom, lottieobj, inkscape_qual=None, force=False):
        n = getattr(lottieobj, "name", None)
        if n is None or self.name_mode == NameMode.NoName:
            if force:
                id = self.gen_id(dom.tag)
                dom.attrib["id"] = id
                return id
            return None

        idn = n.replace(" ", "_")
        if self.id_re.match(idn) and idn not in self.ids:
            self.ids.add(idn)
        else:
            idn = self.gen_id(dom.tag)

        dom.attrib["id"] = idn
        if inkscape_qual:
            dom.attrib[inkscape_qual] = n
        return n

    def _on_animation(self, animation: objects.Animation):
        self.svg.attrib["width"] = str(animation.width)
        self.svg.attrib["height"] = str(animation.height)
        self.svg.attrib["viewBox"] = "0 0 %s %s" % (animation.width, animation.height)
        self.svg.attrib["version"] = "1.1"
        self.set_id(self.svg, animation, self.qualified("sodipodi", "docname"))
        self.defs = ElementTree.SubElement(self.svg, "defs")
        if self.name_mode == NameMode.Inkscape:
            self.svg.attrib[self.qualified("inkscape", "export-xdpi")] = "96"
            self.svg.attrib[self.qualified("inkscape", "export-ydpi")] = "96"
            namedview = ElementTree.SubElement(self.svg, self.qualified("sodipodi", "namedview"))
            namedview.attrib[self.qualified("inkscape", "pagecheckerboard")] = "true"
            namedview.attrib["borderlayer"] = "true"
            namedview.attrib["bordercolor"] = "#666666"
            namedview.attrib["pagecolor"] = "#ffffff"
        self.svg.attrib["style"] = "fill: none; stroke: none"

        return self.svg

    def _on_layer(self, layer_builder, dom_parent):
        if layer_builder.lottie.in_point > self.time or layer_builder.lottie.out_point < self.time:
            return None

        g = self.group_from_lottie(layer_builder.lottie, dom_parent, True)
        if isinstance(layer_builder.lottie, objects.NullLayer):
            g.attrib["opacity"] = "1"
        if not layer_builder.lottie.name:
            g.attrib[self.qualified("inkscape", "label")] = layer_builder.lottie.__class__.__name__
        if layer_builder.shapegroup:
            g.attrib["style"] = self.group_to_style(layer_builder.shapegroup)
        if layer_builder.lottie.hidden:
            g.attrib.setdefault("style", "")
            g.attrib["style"] += "display: none;"

        return g

    def _get_value(self, prop, default=NVector(0, 0)):
        if prop:
            v = prop.get_value(self.time)
        else:
            v = default

        if v is None:
            return default
        if isinstance(v, NVector):
            return v.clone()
        return v

    def set_transform(self, dom, transform):
        trans = []
        pos = self._get_value(transform.position)
        anchor = self._get_value(transform.anchor_point)
        pos -= anchor
        if pos[0] != 0 or pos[1] != 0:
            trans.append("translate(%s, %s)" % (pos.components[0], pos.components[1]))

        scale = self._get_value(transform.scale, NVector(100, 100))
        if scale[0] != 100 or scale[1] != 100:
            scale /= 100
            trans.append("scale(%s, %s)" % (scale.components[0], scale.components[1]))

        rot = self._get_value(transform.rotation, 0)
        if rot != 0:
            trans.append("rotate(%s, %s, %s)" % (rot, anchor[0], anchor[1]))

        if transform.opacity is not None:
            op = transform.opacity.get_value(self.time)
            if op != 100:
                dom.attrib["opacity"] = str(op/100)

        skew = self._get_value(transform.skew, 0)
        if skew != 0:
            axis = self._get_value(transform.skew_axis, 0) * math.pi / 180
            skx = skew * math.cos(axis)
            sky = skew * math.sin(axis)
            # TODO looks like skew moves things around in svg
            trans.append("skewX(%s)" % skx)
            trans.append("skewY(%s)" % sky)

        if trans:
            dom.attrib["transform"] = " ".join(trans)

    def group_to_style(self, group):
        style = {}
        if group.fill:
            style["fill-opacity"] = group.fill.opacity.get_value(self.time) / 100
            if isinstance(group.fill, objects.GradientFill):
                style["fill"] = "url(#%s)" % self.process_gradient(group.fill)
            else:
                style["fill"] = color_to_css(group.fill.color.get_value(self.time))

        if group.stroke:
            if isinstance(group.stroke, objects.GradientStroke):
                style["stroke"] = "url(#%s)" % self.process_gradient(group.stroke)
            else:
                style["stroke"] = color_to_css(group.stroke.color.get_value(self.time))

            style["stroke-opacity"] = group.stroke.opacity.get_value(self.time) / 100
            style["stroke-width"] = group.stroke.width.get_value(self.time)
            if group.stroke.miter_limit is not None:
                style["stroke-miterlimit"] = group.stroke.miter_limit

            if group.stroke.line_cap == objects.LineCap.Round:
                style["stroke-linecap"] = "round"
            elif group.stroke.line_cap == objects.LineCap.Butt:
                style["stroke-linecap"] = "butt"
            elif group.stroke.line_cap == objects.LineCap.Square:
                style["stroke-linecap"] = "square"

            if group.stroke.line_join == objects.LineJoin.Round:
                style["stroke-linejoin"] = "round"
            elif group.stroke.line_join == objects.LineJoin.Bevel:
                style["stroke-linejoin"] = "bevel"
            elif group.stroke.line_join == objects.LineJoin.Miter:
                style["stroke-linejoin"] = "miter"

        return ";".join(map(
            lambda x: ":".join(map(str, x)),
            style.items()
        ))

    def process_gradient(self, gradient):
        spos = gradient.start_point.get_value(self.time)
        epos = gradient.end_point.get_value(self.time)

        if gradient.gradient_type == objects.GradientType.Linear:
            dom = ElementTree.SubElement(self.defs, "linearGradient")
            dom.attrib["x1"] = str(spos[0])
            dom.attrib["y1"] = str(spos[1])
            dom.attrib["x2"] = str(epos[0])
            dom.attrib["y2"] = str(epos[1])
        elif gradient.gradient_type == objects.GradientType.Radial:
            dom = ElementTree.SubElement(self.defs, "radialGradient")
            dom.attrib["cx"] = str(spos[0])
            dom.attrib["cy"] = str(spos[1])
            dom.attrib["r"] = str((epos-spos).length)
            a = gradient.highlight_angle.get_value(self.time) * math.pi / 180
            l = gradient.highlight_length.get_value(self.time)
            dom.attrib["fx"] = str(spos[0] + math.cos(a) * l)
            dom.attrib["fy"] = str(spos[1] + math.sin(a) * l)

        id = self.set_id(dom, gradient, force=True)
        dom.attrib["gradientUnits"] = "userSpaceOnUse"

        colors = gradient.colors.colors.get_value(self.time)
        for i in range(0, len(colors), 4):
            off = colors[i]
            color = colors[i+1:i+4]
            stop = ElementTree.SubElement(dom, "stop")
            stop.attrib["offset"] = "%s%%" % (off * 100)
            stop.attrib["stop-color"] = color_to_css(color)

        return id

    def group_from_lottie(self, lottie, dom_parent, layer):
        g = ElementTree.SubElement(dom_parent, "g")
        if layer and self.name_mode == NameMode.Inkscape:
            g.attrib[self.qualified("inkscape", "groupmode")] = "layer"
        self.set_id(g, lottie, self.qualified("inkscape", "label"), force=True)
        self.set_transform(g, lottie.transform)
        return g

    def _on_shapegroup(self, group, dom_parent):
        if group.empty():
            return

        if len(group.children) == 1 and isinstance(group.children[0], restructure.RestructuredPathMerger):
            path = self.build_path(group.paths.paths, dom_parent)
            self.set_id(path, group.paths.paths[0], force=True)
            path.attrib["style"] = self.group_to_style(group)
            self.set_transform(path, group.lottie.transform)
            return

        g = self.group_from_lottie(group.lottie, dom_parent, group.layer)
        g.attrib["style"] = self.group_to_style(group)
        self.shapegroup_process_children(group, g)
        return g

    def _on_merged_path(self, shape, shapegroup, out_parent):
        path = self.build_path(shape.paths, out_parent)
        self.set_id(path, shape.paths[0])
        path.attrib["style"] = self.group_to_style(shapegroup)
        return path

    def _on_shape(self, shape, shapegroup, out_parent):
        if isinstance(shape, objects.Rect):
            svgshape = self.build_rect(shape, out_parent)
        elif isinstance(shape, objects.Ellipse):
            svgshape = self.build_ellipse(shape, out_parent)
        elif isinstance(shape, objects.Star):
            svgshape = self.build_path([shape.to_bezier()], out_parent)
        elif isinstance(shape, objects.Path):
            svgshape = self.build_path([shape], out_parent)
        elif has_font and isinstance(shape, font.FontShape):
            svgshape = self.build_text(shape, out_parent)
        else:
            return
        self.set_id(svgshape, shape, force=True)
        if "style" not in svgshape.attrib:
            svgshape.attrib["style"] = ""
        svgshape.attrib["style"] += self.group_to_style(shapegroup)
        if shape.hidden:
            svgshape.attrib["style"] += "display: none;"
        return svgshape

    def build_rect(self, shape, parent):
        rect = ElementTree.SubElement(parent, "rect")
        size = shape.size.get_value(self.time)
        pos = shape.position.get_value(self.time)
        rect.attrib["width"] = str(size[0])
        rect.attrib["height"] = str(size[1])
        rect.attrib["x"] = str(pos[0] - size[0] / 2)
        rect.attrib["y"] = str(pos[1] - size[1] / 2)
        rect.attrib["rx"] = str(shape.rounded.get_value(self.time))
        return rect

    def build_ellipse(self, shape, parent):
        ellipse = ElementTree.SubElement(parent, "ellipse")
        size = shape.size.get_value(self.time)
        pos = shape.position.get_value(self.time)
        ellipse.attrib["rx"] = str(size[0] / 2)
        ellipse.attrib["ry"] = str(size[1] / 2)
        ellipse.attrib["cx"] = str(pos[0])
        ellipse.attrib["cy"] = str(pos[1])
        return ellipse

    def build_path(self, shapes, parent):
        path = ElementTree.SubElement(parent, "path")
        d = ""
        for shape in shapes:
            bez = shape.shape.get_value(self.time)
            if not bez.vertices:
                continue
            if d:
                d += "\n"
            d += "M %s,%s " % tuple(bez.vertices[0].components[:2])
            for i in range(1, len(bez.vertices)):
                qfrom = bez.vertices[i-1]
                h1 = bez.out_tangents[i-1] + qfrom
                qto = bez.vertices[i]
                h2 = bez.in_tangents[i] + qto
                d += "C %s,%s %s,%s %s,%s " % (
                    h1[0], h1[1],
                    h2[0], h2[1],
                    qto[0], qto[1],
                )
            if bez.closed:
                qfrom = bez.vertices[-1]
                h1 = bez.out_tangents[-1] + qfrom
                qto = bez.vertices[0]
                h2 = bez.in_tangents[0] + qto
                d += "C %s,%s %s,%s %s,%s Z" % (
                    h1[0], h1[1],
                    h2[0], h2[1],
                    qto[0], qto[1],
                )

        path.attrib["d"] = d
        return path

    def _on_shape_modifier(self, shape, shapegroup, out_parent):
        if isinstance(shape.lottie, objects.Repeater):
            svgshape = self.build_repeater(shape.lottie, shape.child, shapegroup, out_parent)
        elif isinstance(shape.lottie, objects.RoundedCorners):
            svgshape = self.build_rouded_corners(shape.lottie, shape.child, shapegroup, out_parent)
        else:
            return self.shapegroup_process_child(shape.child, shapegroup, out_parent)
        if svgshape:
            self.set_id(svgshape, shape, force=True)
            svgshape.attrib["style"] = self.group_to_style(shapegroup)
        return svgshape

    def build_repeater(self, shape, child, shapegroup, out_parent):
        original = self.shapegroup_process_child(child, shapegroup, out_parent)
        if not original:
            return

        ncopies = int(round(shape.copies.get_value(self.time)))
        if ncopies == 1:
            return

        out_parent.remove(original)

        g = ElementTree.SubElement(out_parent, "g")

        for copy in range(ncopies-1):
            use = ElementTree.SubElement(g, "use")
            use.attrib[self.qualified("xlink", "href")] = "#" + original.attrib["id"]

        orig_wrapper = ElementTree.SubElement(g, "g")
        orig_wrapper.append(original)

        transform = objects.Transform()
        so = shape.transform.start_opacity.get_value(self.time)
        eo = shape.transform.end_opacity.get_value(self.time)
        position = shape.transform.position.get_value(self.time)
        rotation = shape.transform.rotation.get_value(self.time)
        anchor_point = shape.transform.anchor_point.get_value(self.time)
        for i in range(ncopies-1, -1, -1):
            of = i / (ncopies-1)
            transform.opacity.value = so * of + eo * (1 - of)
            self.set_transform(g[i], transform)
            transform.position.value += position
            transform.rotation.value += rotation
            transform.anchor_point.value += anchor_point

        return g

    def build_rouded_corners(self, shape, child, shapegroup, out_parent):
        round_amount = shape.radius.get_value(self.time)
        if isinstance(child, objects.Shape):
            path = child.to_bezier()
            bezier = path.shape.get_value(self.time).rounded(round_amount)
            path.shape.clear_animation(bezier)
            return self._on_shape(path, shapegroup, out_parent)

        if isinstance(child, restructure.RestructuredShapeGroup):
            self._build_rouded_corners_group(child, round_amount)
            return self._on_shapegroup(child, out_parent)

        return child

    def _build_rouded_corners_group(self, shapegroup, round_amount):
        children = []
        for sh in shapegroup.children:
            if isinstance(sh, objects.Shape):
                path = sh.to_bezier()
                bezier = path.shape.get_value(self.time).rounded(round_amount)
                path.shape.clear_animation(bezier)
                sh = path
            elif isinstance(sh, restructure.RestructuredShapeGroup):
                self._build_rouded_corners_group(sh, round_amount)
            children.append(sh)
        shapegroup.children = children

    def _custom_object_supported(self, shape):
        if has_font and isinstance(shape, font.FontShape):
            return True
        return False

    def build_text(self, shape, parent):
        text = ElementTree.SubElement(parent, "text")
        if "family" in shape.query:
            text.attrib["font-family"] = shape.query["family"]
        if "weight" in shape.query:
            text.attrib["font-weight"] = str(shape.query.weight_to_css())
        slant = int(shape.query.get("slant", 0))
        if slant > 0 and slant < 110:
            text.attrib["font-style"] = "italic"
        elif slant >= 110:
            text.attrib["font-style"] = "oblique"

        text.attrib["font-size"] = str(shape.size)

        text.attrib["white-space"] = "pre"

        pos = shape.style.position
        text.attrib["x"] = str(pos.x)
        text.attrib["y"] = str(pos.y)
        text.text = shape.text

        return text


def color_to_css(color):
    #if len(color) == 4:
        #return ("rgba(%s, %s, %s" % tuple(map(lambda c: int(round(c*255)), color[:3]))) + ", %s)" % color[3]
    return "rgb(%s, %s, %s)" % tuple(map(lambda c: int(round(c*255)), color[:3]))


def to_svg(animation, time):
    builder = SvgBuilder(time)
    builder.process(animation)
    return builder.dom



from xml.dom import minidom
from xml.etree import ElementTree

from .base import exporter
from ..parsers.svg.builder import to_svg


def _print_ugly_xml(dom, fp):
    return dom.write(fp, "utf-8", True)


def _print_pretty_xml(dom, fp):
    xmlstr = minidom.parseString(ElementTree.tostring(dom.getroot())).toprettyxml(indent="   ")
    if isinstance(fp, str):
        fp = open(fp, "w")
    fp.write(xmlstr)


@exporter("SVG", ["svg"], [], {"pretty", "frame"})
def export_svg(animation, fp, frame=0, pretty=True):
    _print_xml = _print_pretty_xml if pretty else _print_ugly_xml
    _print_xml(to_svg(animation, frame), fp)



from .base import TgsObject, TgsProp, PseudoBool
from .layers import Layer
from .shapes import ShapeElement


## \ingroup Lottie
class Asset(TgsObject):
    @classmethod
    def _load_get_class(cls, lottiedict):
        if "p" in lottiedict or "u" in lottiedict:
            return Image
        if "layers" in lottiedict:
            return Precomp


## \ingroup Lottie
class Image(Asset):
    """!
        External image

        \see http://docs.aenhancers.com/sources/filesource/
    """
    _props = [
        TgsProp("height", "h", float, False),
        TgsProp("width", "w", float, False),
        TgsProp("id", "id", str, False),
        TgsProp("image", "p", str, False),
        TgsProp("image_path", "u", str, False),
        TgsProp("embedded", "e", PseudoBool, False),
    ]

    def __init__(self, id=""):
        ## Image Height
        self.height = 0
        ## Image Width
        self.width = 0
        ## Image ID
        self.id = id
        ## Image name
        self.image = ""
        ## Image path
        self.image_path = ""
        ## Image data is stored as a data: url
        self.embedded = False

    def load(self, file, format="png"):
        """!
        \param file     Filename or file object to load
        \param format   Format to store the image data as
        """
        from PIL import Image
        from io import BytesIO
        import base64
        import os
        im = Image.open(file)
        self.width, self.height = im.size
        output = BytesIO()
        im.save(output, format=format)
        self.image = "data:image/%s;base64,%s" % (
            format,
            base64.b64encode(output.getvalue()).decode("ascii")
        )
        self.embedded = True
        if not self.id:
            if isinstance(file, str):
                self.id = os.path.basename(file)
            elif hasattr(file, "name"):
                self.id = os.path.basename(file.name)
            else:
                self.id = "image_%s" % id(self)
        return self


## \ingroup Lottie
class CharacterData(TgsObject):
    """!
    Character shapes
    """
    _props = [
        TgsProp("shapes", "shapes", ShapeElement, True),
    ]

    def __init__(self):
        self.shapes = []


## \ingroup Lottie
class Chars(TgsObject):
    """!
    Defines character shapes to avoid loading system fonts
    """
    _props = [
        TgsProp("character", "ch", str, False),
        TgsProp("font_family", "fFamily", str, False),
        TgsProp("font_size", "size", float, False),
        TgsProp("font_style", "style", str, False),
        TgsProp("width", "w", float, False),
        TgsProp("data", "data", CharacterData, False),
    ]

    def __init__(self):
        ## Character Value
        self.character = ""
        ## Character Font Family
        self.font_family = ""
        ## Character Font Size
        self.font_size = 0
        ## Character Font Style
        self.font_style = "" # Regular
        ## Character Width
        self.width = 0
        ## Character Data
        self.data = CharacterData()

    @property
    def shapes(self):
        return self.data.shapes


## \ingroup Lottie
## \ingroup LottieCheck
class Precomp(Asset):
    _props = [
        TgsProp("id", "id", str, False),
        TgsProp("layers", "layers", Layer, True),
    ]

    def __init__(self):
        ## Precomp ID
        self.id = ""
        ## List of Precomp Layers
        self.layers = []



import sys

from ..objects.base import TgsObject, Tgs
from ..objects.properties import MultiDimensional, Value, ShapeProperty
from ..objects.layers import Layer


def _prettyprint_scalar(tgs_object, out=sys.stdout):
    if isinstance(tgs_object, float) and tgs_object == round(tgs_object):
        tgs_object = int(tgs_object)
    return str(tgs_object)


def prettyprint(tgs_object, out=sys.stdout, indent="   ", _i=""):
    if isinstance(tgs_object, TgsObject):
        out.write(tgs_object.__class__.__name__)
        out.write('\n')
        _i += indent
        maxk = max(map(lambda x: len(x.name), tgs_object._props))
        for k in tgs_object._props:
            out.write(_i)
            out.write(k.name.ljust(maxk))
            out.write(' : ')
            prettyprint(k.get(tgs_object), out, indent, _i)
    elif isinstance(tgs_object, (list, tuple)):
        if not tgs_object or (not isinstance(tgs_object[0], Tgs) and len(tgs_object) < 16):
            out.write("[")
            out.write(", ".join(map(_prettyprint_scalar, tgs_object)))
            out.write("]\n")
        else:
            out.write("[\n")
            for k in tgs_object:
                out.write(_i + indent)
                prettyprint(k, out, indent, _i + indent)
            out.write(_i)
            out.write(']\n')
    else:
        out.write(_prettyprint_scalar(tgs_object, out))
        out.write('\n')


def _prettyprint_summary_printable(obj):
    if isinstance(obj, TgsObject):
        return not isinstance(obj, (MultiDimensional, Value, ShapeProperty))
    return obj and isinstance(obj, (list, tuple)) and isinstance(obj[0], TgsObject)


def prettyprint_summary(tgs_object, out=sys.stdout, indent="   ", _i=""):
    if isinstance(tgs_object, TgsObject):
        out.write(tgs_object.__class__.__name__)
        name = getattr(tgs_object, "name", None)
        if name:
            out.write(" %r" % name)
        if isinstance(tgs_object, Layer):
            out.write(" %s -> %s" % (tgs_object.index, tgs_object.parent))
        out.write('\n')
        _i += indent
        for k in tgs_object._props:
            val = k.get(tgs_object)
            if _prettyprint_summary_printable(val):
                out.write(_i)
                out.write(k.name)
                out.write(' : ')
                prettyprint_summary(val, out, indent, _i)
    elif _prettyprint_summary_printable(tgs_object):
            out.write("[\n")
            for k in tgs_object:
                out.write(_i + indent)
                prettyprint_summary(k, out, indent, _i + indent)
            out.write(_i)
            out.write(']\n')



import os
import pkgutil
import importlib


class Exporter:
    def __init__(self, name, extensions, callback, extra_options=[], generic_options=set(), slug=None):
        self.name = name
        self.extensions = extensions
        self.callback = callback
        self.extra_options = extra_options
        self.generic_options = generic_options
        self.slug = slug if slug is not None else extensions[0]

    def export(self, animation, filename, options):
        self.callback(animation, filename, **options)

    def __repr__(self):
        return "<Exporter %s>" % self.slug

    def argparse_options(self, ns):
        o_options = {}
        for opt in self.extra_options:
            o_options[opt.name] = getattr(ns, opt.nsvar(self.slug))
        for opt in self.generic_options:
            o_options[opt] = getattr(ns, opt)
        return o_options


class ExtraOption:
    def __init__(self, name, **kwargs):
        self.name = name
        self.kwargs = kwargs

    def add_argument(self, slug, parser):
        opt = "--%s-%s" % (slug, self.name.replace("_", "-"))
        parser.add_argument(opt, metavar=self.name, **self.kwargs)

    def nsvar(self, slug):
        return "%s_%s" % (slug, self.name)


_exporters = {}


def exporter(name, extensions, extra_options=[], generic_options=set(), slug=None):
    def decorator(callback):
        exporter = Exporter(name, extensions, callback, extra_options, generic_options, slug)
        _exporters[exporter.slug] = exporter
        return callback
    return decorator


def _add_options(parser, ie, object):
    if not object.extra_options:
        return

    suf = " %sing options" % ie
    group = parser.add_argument_group(object.name + suf)
    for op in object.extra_options:
        op.add_argument(object.slug, group)


class _ExporterLoader:
    def __init__(self):
        self._exporters = None

    def load_modules(self):
        for _, modname, _ in pkgutil.iter_modules([os.path.dirname(__file__)]):
            if modname == "base":
                continue

            #full_modname = "tgs.exporters." + modname
            full_modname = "." + modname
            try:
                importlib.import_module(full_modname, "tgs.exporters")
            except ImportError:
                pass

    @property
    def exporters(self):
        if self._exporters is None:
            self.load_modules()
            self._exporters = _exporters
        return self._exporters

    def __iter__(self):
        return iter(self.exporters.values())

    def get(self, slug):
        return self.exporters.get(slug, None)

    def __getitem__(self, key):
        return self.get(key)

    def get_from_filename(self, filename):
        return self.get_from_extension(os.path.splitext(filename)[1][1:])

    def get_from_extension(self, ext):
        for p in self.exporters.values():
            if ext in p.extensions:
                return p
        return None

    def set_options(self, parser):
        group = parser.add_argument_group("Generic output options")
        group.add_argument(
            "--pretty", "-p",
            action="store_true",
            help="Pretty print (for formats that support it)",
        )
        group.add_argument(
            "--frame",
            type=int,
            default=0,
            help="Frame to extract (for single-image formats)",
        )

        for exporter in self.exporters.values():
            _add_options(parser, "export", exporter)

        return group

    def keys(self):
        return self.exporters.keys()


exporters = _ExporterLoader()



from .base import TgsObject, TgsProp, PseudoBool, Index
from .layers import Layer
from .assets import Asset, Chars
from .text import FontList

##\defgroup Lottie Lottie
#
# Objects of the lottie file structure.

## \defgroup LottieCheck Lottie (to check)
#
# Lottie objects that have not been tested


## \ingroup Lottie
class Animation(TgsObject):
    """!
    Top level object, describing the animation

    @see http://docs.aenhancers.com/items/compitem/
    @todo rename to Composition?
    """
    _props = [
        TgsProp("version", "v", str, False),
        TgsProp("frame_rate", "fr", float, False),
        TgsProp("in_point", "ip", float, False),
        TgsProp("out_point", "op", float, False),
        TgsProp("width", "w", int, False),
        TgsProp("height", "h", int, False),
        TgsProp("name", "nm", str, False),
        TgsProp("threedimensional", "ddd", PseudoBool, False),
        TgsProp("assets", "assets", Asset, True),
        #TgsProp("comps", "comps", Animation, True),
        TgsProp("fonts", "fonts", FontList),
        TgsProp("layers", "layers", Layer, True),
        TgsProp("chars", "chars", Chars, True),
        #TgsProp("markers", "markers", Marker, True),
        TgsProp("tgs", "tgs", PseudoBool, False),
        #TgsProp("motion_blur", "mb", MotionBlur, False),
    ]
    _version = "5.5.2"

    def __init__(self, n_frames=60, framerate=60):
        ## Marks as telegram sticker
        self.tgs = 1
        ## The time when the composition work area begins, in frames.
        self.in_point = 0
        ## The time when the composition work area ends.
        ## Sets the final Frame of the animation
        self.out_point = n_frames
        ## Frames per second
        self.frame_rate = framerate
        ## Composition Width
        self.width = 512
        ## Composition has 3-D layers
        self.threedimensional = False
        ## Composition Height
        self.height = 512
        ## Bodymovin Version
        self.version = self._version
        ## Composition name
        self.name = None
        ## List of Composition Layers
        self.layers = [] # ShapeLayer, SolidLayer, CompLayer, ImageLayer, NullLayer, TextLayer
        ## source items that can be used in multiple places. Comps and Images for now.
        self.assets = [] # Image, Precomp
        ## source chars for text layers
        self.chars = None
        ## Available fonts
        self.fonts = None
        self._index_gen = Index()

    def add_layer(self, layer):
        """!
        @brief Appends a layer to the animation
        \see insert_layer
        """
        return self.insert_layer(len(self.layers), layer)

    def insert_layer(self, index, layer):
        """!
        @brief Inserts a layer to the animation
        @note Layers added first will be rendered on top of later layers
        """
        self.layers.insert(index, layer)
        if layer.index is None:
            layer.index = next(self._index_gen)
        if layer.in_point is None:
            layer.in_point = self.in_point
        if layer.out_point is None:
            layer.out_point = self.out_point
        return layer

    def clone(self):
        c = super().clone()
        c._index_gen._i = self._index_gen._i
        return c

    def tgs_sanitize(self):
        """!
        Cleans up some things to ensure it works as a telegram sticker
        """
        if self.width != 512 or self.height != 512:
            scale = min(512/self.width, 512/self.height)
            self.width = self.height = 512

            for layer in self.layers:
                if layer.transform.scale.animated:
                    for kf in layer.transform.scale.keyframes:
                        if kf.start is not None:
                            kf.start *= scale
                        if kf.end is not None:
                            kf.end *= scale
                else:
                    layer.transform.scale.value *= scale

        if self.frame_rate < 45:
            self.frame_rate = 30
        else:
            self.frame_rate = 60



import math
from xml.dom import minidom

from ... import objects
from ...utils import restructure


blend_modes = {
    objects.BlendMode.Normal: 0, # composite
    objects.BlendMode.Multiply: 6,
    objects.BlendMode.Screen: 16,
    objects.BlendMode.Overlay: 20,
    objects.BlendMode.Darken: 3,
    objects.BlendMode.Lighten: 2,
    objects.BlendMode.HardLight: 17,
    objects.BlendMode.Difference: 18,
    objects.BlendMode.Hue: 9,
    objects.BlendMode.Saturation: 10,
    objects.BlendMode.Color: 8,
    objects.BlendMode.Luminosity: 11,
    #1  : "straight",
    #13 : "onto",
    #21 : "straight onto",
    #12 : "behind",
    #7  : "divide",
    #4  : "add",
    #5  : "subtract",
    objects.BlendMode.Exclusion: 18,
    objects.BlendMode.SoftLight: 6,
    objects.BlendMode.ColorDodge: 0,
    objects.BlendMode.ColorBurn: 0,
}


class SifBuilder(restructure.AbstractBuilder):
    def __init__(self, gamma=2.2):
        """
        @todo Add gamma option to tgsconvert
        """
        super().__init__()
        self.dom = minidom.Document()
        self.canvas = self.dom.appendChild(self.dom.createElement("canvas"))
        self.canvas.setAttribute("version", "1.0")
        self.gamma = gamma
        self.autoid = objects.base.Index()

    def _format_time_f(self, time):
        return "%sf" % time

    def _format_time(self, time):
        return "%ss" % (time / self.framerate)

    def _subelement(self, parent, tagname):
        return parent.appendChild(self.dom.createElement(tagname))

    def _settext(self, element, text):
        element.appendChild(self.dom.createTextNode(text))

    def _on_animation(self, animation: objects.Animation):
        if animation.name:
            self._settext(self._subelement(self.canvas, "name"), animation.name)
        self.canvas.setAttribute("width", str(animation.width))
        self.canvas.setAttribute("height", str(animation.height))
        self.canvas.setAttribute("xres", str(animation.width))
        self.canvas.setAttribute("yres", str(animation.height))
        self.canvas.setAttribute("view-box", "0 0 %s %s" % (animation.width, animation.height))
        self.canvas.setAttribute("fps", str(animation.frame_rate))
        self.canvas.setAttribute("begin-time", self._format_time_f(animation.in_point))
        self.canvas.setAttribute("end-time", self._format_time_f(animation.out_point))
        self.canvas.setAttribute("antialias", "1")
        self.framerate = animation.frame_rate
        self.end_frame = animation.out_point
        self.width = animation.width
        self.height = animation.height
        self.defs = self._subelement(self.canvas, "defs")
        return self.canvas

    def _on_layer(self, layer_builder, dom_parent):
        layer = self.layer_from_lottie("group", layer_builder.lottie, dom_parent)
        if not layer_builder.lottie.name:
            layer.setAttribute("desc", layer_builder.lottie.__class__.__name__)

        bm = getattr(layer_builder.lottie, "blend_mode", objects.BlendMode.Normal)
        self.simple_param("blend_method", bm, layer, "integer").setAttribute("static", "true")

        stretch = getattr(layer_builder.lottie, "stretch", 1) or 1
        self.simple_param("time_dilation", stretch, layer)

        in_point = getattr(layer_builder.lottie, "in_point", 0)
        self.simple_param("time_offset", self._format_time(in_point), layer, "time")

        pcanv = self._subelement(layer, "param")
        pcanv.setAttribute("name", "canvas")
        canvas = self._subelement(pcanv, "canvas")
        out_point = getattr(layer_builder.lottie, "out_point", self.end_frame)
        canvas.setAttribute("end-time", self._format_time(out_point))
        return canvas

    def basic_layer(self, type, dom_parent):
        g = self._subelement(dom_parent, "layer")
        g.setAttribute("type", type)
        g.setAttribute("active", "true")
        g.setAttribute("exclude_from_rendering", "false")
        #g.setAttribute("version", "0.2")
        return g

    def layer_from_lottie(self, type, lottie, dom_parent):
        g = self.basic_layer(type, dom_parent)
        if lottie.name:
            g.setAttribute("desc", lottie.name)
        if lottie.hidden:
            g.setAttribute("active", "false")
        transf = getattr(lottie, "transform", None)
        if transf:
            self.set_transform(g, lottie.transform)
        return g

    def _get_scale(self, transform):
        def func(keyframe, elem):
            t = keyframe.time if keyframe else 0
            scale_x, scale_y = transform.scale.get_value(t)[:2]
            scale_x /= 100
            scale_y /= 100
            skew = transform.skew.get_value(t) if transform.skew else 0
            c = math.cos(skew * math.pi / 180)
            if c != 0:
                scale_y *= 1 / c
            self._settext(self._subelement(elem, "x"), str(scale_x))
            self._settext(self._subelement(elem, "y"), str(scale_y))
        return func

    def set_transform(self, group, transform):
        param = self._subelement(group, "param")
        param.setAttribute("name", "transformation")
        composite = self._subelement(param, "composite")
        composite.setAttribute("type", "transformation")
        self.process_vector("offset", transform.position, composite)

        keyframes = self._merge_keyframes([transform.scale, transform.skew])

        self.process_vector_ext("scale", keyframes, composite, "vector", self._get_scale(transform))
        #self.process_vector_ext("scale", transform.scale.keyframes, composite, "vector", get_scale)
        self.process_scalar("angle", "skew_angle", transform.skew or objects.Value(0), composite)
        self.process_scalar("angle", "angle", transform.rotation, composite)
        self.process_scalar("real", "param", transform.opacity, group, 1/100).setAttribute("name", "amount")
        self.process_vector("param", transform.anchor_point, group).setAttribute("name", "origin")
        # TODO get z_depth from position
        self.process_scalar("real", "param", objects.Value(0), group).setAttribute("name", "z_depth")

    def process_vector(self, name, multidim, parent):
        def getter(keyframe, elem):
            if keyframe is None:
                v = multidim.value
            else:
                v = keyframe.start
            self._settext(self._subelement(elem, "x"), str(v[0]))
            self._settext(self._subelement(elem, "y"), str(v[1]))

        return self.process_vector_ext(name, multidim.keyframes, parent, "vector", getter)

    def process_vector(self, name, multidim, parent):
        def getter(keyframe, elem):
            if keyframe is None:
                v = multidim.value
            else:
                v = keyframe.start
            self._settext(self._subelement(elem, "x"), str(v[0]))
            self._settext(self._subelement(elem, "y"), str(v[1]))

        return self.process_vector_ext(name, multidim.keyframes, parent, "vector", getter)

    def process_vector_ext(self, name, kframes, parent, type, getter):
        wrap = self._subelement(parent, name)
        if kframes is not None:
            animated = self._subelement(wrap, "animated")
            animated.setAttribute("type", type)
            for i in range(len(kframes)):
                keyframe = kframes[i]
                waypoint = self._subelement(animated, "waypoint")
                waypoint.setAttribute("time", self._format_time(keyframe.time))

                if i > 0:
                    prev = kframes[i-1]
                    if prev.jump:
                        waypoint.setAttribute("before", "constant")
                    elif prev.in_value and prev.in_value.x < 1:
                        waypoint.setAttribute("before", "halt")
                    else:
                        waypoint.setAttribute("before", "linear")
                else:
                    waypoint.setAttribute("before", "linear")

                if keyframe.jump:
                    waypoint.setAttribute("after", "constant")
                elif keyframe.out_value and keyframe.out_value.x > 0:
                    waypoint.setAttribute("after", "halt")
                else:
                    waypoint.setAttribute("after", "linear")

                vector = self._subelement(waypoint, type)
                getter(keyframe, vector)
        else:
            vector = self._subelement(wrap, type)
            getter(None, vector)
        return wrap

    def process_scalar(self, type, name, value, parent, mult=None):
        def getter(keyframe, elem):
            if keyframe is None:
                v = value.value
            else:
                v = keyframe.start[0]
            if mult is not None:
                v *= mult
            elem.setAttribute("value", str(v))
        return self.process_vector_ext(name, value.keyframes, parent, type, getter)

    def simple_param(self, name, vaue, parent, type="real"):
        param = self._subelement(parent, "param")
        param.setAttribute("name", name)
        e_val = self._subelement(param, type)
        e_val.setAttribute("value", str(vaue))
        return e_val

    def simple_composite_param(self, name, vaue, parent, type="real"):
        param = self._subelement(parent, name)
        e_val = self._subelement(param, type)
        e_val.setAttribute("value", str(vaue))
        return e_val

    def _on_shape(self, shape, group, dom_parent):
        layers = []
        if not hasattr(shape, "to_bezier"):
            return []

        if group.fill:
            sif_shape = self.build_path("region", shape.to_bezier(), dom_parent)
            layers.append(sif_shape)
            self.apply_group_fill(sif_shape, group.fill)

        if group.stroke:
            sif_shape = self.build_path("outline", shape.to_bezier(), dom_parent)
            self.apply_group_stroke(sif_shape, group.stroke)
            layers.append(sif_shape)

        return layers

    def _merge_keyframes(self, props):
        keyframes = {}
        for prop in props:
            if prop is not None and prop.animated:
                keyframes.update({kf.time: kf for kf in prop.keyframes})
        return list(sorted(keyframes.values(), key=lambda kf: kf.time)) or None

    def apply_group_fill(self, sif_shape, fill):
        ## @todo gradients?
        if hasattr(fill, "colors"):
            return

        def getter(keyframe, elem):
            if keyframe is None:
                v = fill.color.value
            else:
                v = keyframe.start
            self._settext(self._subelement(elem, "r"), str(v[0] ** self.gamma))
            self._settext(self._subelement(elem, "g"), str(v[1] ** self.gamma))
            self._settext(self._subelement(elem, "b"), str(v[2] ** self.gamma))
            self._settext(self._subelement(elem, "a"), "1")

        self.process_vector_ext("param", fill.color.keyframes, sif_shape, "color", getter).setAttribute("name", "color")

        def get_op(keyframe, elem):
            if keyframe is None:
                v = fill.opacity.value
            else:
                v = keyframe.start
            v /= 100
            elem.setAttribute("value", str(v))

        self.process_vector_ext("param", fill.opacity.keyframes, sif_shape, "real", get_op).setAttribute("name", "amount")

    def apply_group_stroke(self, sif_shape, stroke):
        self.apply_group_fill(sif_shape, stroke)
        cusp = self._format_bool(stroke.line_join == objects.LineJoin.Miter)
        self.simple_param("sharp_cusps", cusp, sif_shape, "bool")
        round_cap = self._format_bool(stroke.line_cap == objects.LineCap.Round)
        self.simple_param("round_tip[0]", round_cap, sif_shape, "bool")
        self.simple_param("round_tip[1]", round_cap, sif_shape, "bool")
        self.process_scalar("real", "param", stroke.width, sif_shape).setAttribute("name", "width")

    def _format_bool(self, value):
        return str(bool(value)).lower()

    def build_path(self, type, path, dom_parent):
        layer = self.layer_from_lottie(type, path, dom_parent)
        bline_par = self._subelement(layer, "param")
        bline_par.setAttribute("name", "bline")
        bline = self._subelement(bline_par, "bline")
        bline.setAttribute("type", "bline_point")
        startbez = path.shape.get_value()
        bline.setAttribute("loop", self._format_bool(startbez.closed))
        nverts = len(startbez.vertices)
        for point in range(nverts):
            self.bezier_point(path, point, bline)
        return layer

    def bezier_point(self, lottie_path, point_index, sif_parent):
        entry = self._subelement(sif_parent, "entry")
        composite = self._subelement(entry, "composite")
        composite.setAttribute("type", "bline_point")

        def get_point(keyframe, elem):
            if keyframe is None:
                bezier = lottie_path.shape.value
            else:
                bezier = keyframe.start
            if not bezier:
                elem.parentNode.parentNode.removeChild(elem.parentNode)
                return
            vert = bezier.vertices[point_index]
            self._settext(self._subelement(elem, "x"), str(vert[0]))
            self._settext(self._subelement(elem, "y"), str(vert[1]))

        self.process_vector_ext("point", lottie_path.shape.keyframes, composite, "vector", get_point)
        self.simple_composite_param("split", "true", composite, "bool")
        self.simple_composite_param("split_radius", "true", composite, "bool")
        self.simple_composite_param("split_angle", "true", composite, "bool")
        self.simple_composite_param("width", "1.0", composite, "real")
        self.simple_composite_param("origin", "0.5", composite, "real")

        def get_tangent_r(keyframe, elem):
            if keyframe is None:
                bezier = lottie_path.shape.value
            else:
                bezier = keyframe.start
            if not bezier:
                elem.parentNode.parentNode.removeChild(elem.parentNode)
                return

            inp = getattr(bezier, which_point)[point_index]
            radius = math.hypot(inp.x, inp.y) * 3 * mult
            elem.setAttribute("value", str(radius))

        def get_tangent_th(keyframe, elem):
            if keyframe is None:
                bezier = lottie_path.shape.value
            else:
                bezier = keyframe.start
            if not bezier:
                elem.parentNode.parentNode.removeChild(elem.parentNode)
                return

            inp = getattr(bezier, which_point)[point_index]
            if inp[0] == 0:
                theta = 0
            else:
                theta = math.atan(inp[1] / inp[0]) * 180 / math.pi
                if inp[0] < 0:
                    theta += 180
            elem.setAttribute("value", str(theta))

        mult = -1
        which_point = "in_tangents"
        radial_composite = self._subelement(self._subelement(composite, "t1"), "radial_composite")
        radial_composite.setAttribute("type", "vector")
        self.process_vector_ext("radius", lottie_path.shape.keyframes, radial_composite, "real", get_tangent_r)
        self.process_vector_ext("theta", lottie_path.shape.keyframes, radial_composite, "angle", get_tangent_th)

        mult = 1
        which_point = "out_tangents"
        radial_composite = self._subelement(self._subelement(composite, "t2"), "radial_composite")
        radial_composite.setAttribute("type", "vector")
        self.process_vector_ext("radius", lottie_path.shape.keyframes, radial_composite, "real", get_tangent_r)
        self.process_vector_ext("theta", lottie_path.shape.keyframes, radial_composite, "angle", get_tangent_th)

    def _on_shapegroup(self, shape_group, dom_parent):
        if shape_group.empty():
            return

        layer = self.layer_from_lottie("group", shape_group.lottie, dom_parent)

        pcanv = self._subelement(layer, "param")
        pcanv.setAttribute("name", "canvas")
        canvas = self._subelement(pcanv, "canvas")
        self.shapegroup_process_children(shape_group, canvas)

    def _modifier_inner_group(self, modifier, shapegroup, dom_parent):
        layer = self.basic_layer("group", dom_parent)
        pcanv = self._subelement(layer, "param")
        pcanv.setAttribute("name", "canvas")
        canvas = self._subelement(pcanv, "canvas")
        self.shapegroup_process_child(modifier.child, shapegroup, canvas)
        return layer

    def _on_shape_modifier(self, modifier, shapegroup, dom_parent):
        layer = self.basic_layer("group", dom_parent)
        if modifier.lottie.name:
            layer.setAttribute("desc", modifier.lottie.name)

        pcanv = self._subelement(layer, "param")
        pcanv.setAttribute("name", "canvas")
        canvas = self._subelement(pcanv, "canvas")
        inner = self._modifier_inner_group(modifier, shapegroup, canvas)
        if isinstance(modifier.lottie, objects.Repeater):
            self.build_repeater(modifier.lottie, inner, canvas)

    def _build_repeater_defs(self, shape, name_id):
        dup = self._subelement(self.defs, "duplicate")
        dup.setAttribute("type", "real")
        dup.setAttribute("id", name_id)

        def getter(keyframe, elem):
            if keyframe is None:
                v = shape.copies.value
            else:
                v = keyframe.start[0]
            elem.setAttribute("value", str(v-1))
        self.process_vector_ext("from", shape.copies.keyframes, dup, "real", getter)
        self._subelement(self._subelement(dup, "to"), "real").setAttribute("value", "0")
        self._subelement(self._subelement(dup, "step"), "real").setAttribute("value", "-1")

    def _build_repeater_transform_scale_component(self, shape, name_id, comp, scalecomposite):
        x = self._subelement(scalecomposite, "xy"[comp])
        power = self._subelement(x, "power")
        power.setAttribute("type", "real")
        #power.setAttribute("power", ":" + name_id)

        def getter(keyframe, elem):
            if keyframe is None:
                v = shape.transform.scale.value
            else:
                v = keyframe.start
            v = v[comp] / 100
            elem.setAttribute("value", str(v))

        self.process_vector_ext("base", shape.transform.scale.keyframes, power, "real", getter)
        self.process_scalar("real", "epsilon", objects.Value(0.000001), power)
        self.process_scalar("real", "infinite", objects.Value(999999), power)

        # HACK work around an issue in Synfig
        exponent = self._subelement(power, "power")
        add = self._subelement(exponent, "add")
        add.setAttribute("type", "real")
        add.setAttribute("lhs", ":" + name_id)
        self.process_scalar("real", "rhs", objects.Value(0.000001), add)
        self.process_scalar("real", "scalar", objects.Value(1), add)

    def _build_repeater_transform(self, shape, inner, name_id):
        param = self._subelement(inner, "param")
        param.setAttribute("name", "transformation")
        composite = self._subelement(param, "composite")
        composite.setAttribute("type", "transformation")

        offset = self._subelement(composite, "offset")
        scale = self._subelement(offset, "scale")
        scale.setAttribute("type", "vector")
        scale.setAttribute("scalar", ":" + name_id)
        self.process_vector("link", shape.transform.position, scale)

        angle = self._subelement(composite, "angle")
        scale = self._subelement(angle, "scale")
        scale.setAttribute("type", "angle")
        scale.setAttribute("scalar", ":" + name_id)
        self.process_scalar("angle", "link", shape.transform.rotation, scale)

        scale = self._subelement(composite, "scale")
        scalecomposite = self._subelement(scale, "composite")
        scalecomposite.setAttribute("type", "vector")
        self._build_repeater_transform_scale_component(shape, name_id, 0, scalecomposite)
        self._build_repeater_transform_scale_component(shape, name_id, 1, scalecomposite)
        #self.process_vector_ext(
            #"scale", shape.transform.scale.keyframes, composite,
            #"vector", self._get_scale(shape.transform)
        #)

        self.process_scalar("angle", "skew_angle", objects.Value(0), composite)

    def _build_repeater_amount(self, shape, inner, name_id):
        param = self._subelement(inner, "param")
        param.setAttribute("name", "amount")
        subtract = self._subelement(param, "subtract")
        subtract.setAttribute("type", "real")
        self.process_scalar("real", "scalar", objects.Value(1), subtract)

        self.process_scalar("real", "lhs", shape.transform.start_opacity, subtract, 0.01)

        rhs = self._subelement(subtract, "rhs")
        scale = self._subelement(rhs, "scale")
        scale.setAttribute("type", "real")
        scale.setAttribute("scalar", ":" + name_id)

        def getter(keyframe, elem):
            if keyframe is None:
                t = 0
                end = shape.transform.end_opacity.value
            else:
                t = keyframe.time
                end = keyframe.start[0]
            start = shape.transform.start_opacity.get_value(t)
            n = shape.copies.get_value(t)
            v = (start - end) / (n - 1) / 100 if n > 0 else 0
            elem.setAttribute("value", str(v))
        self.process_vector_ext("link", shape.transform.end_opacity.keyframes, scale, "real", getter)

    def build_repeater(self, shape, inner, dom_parent):
        name_id = "duplicate_%s" % next(self.autoid)
        self._build_repeater_defs(shape, name_id)
        self._build_repeater_transform(shape, inner, name_id)
        self._build_repeater_amount(shape, inner, name_id)

        # duplicate layer
        duplicate = self.basic_layer("duplicate", dom_parent)
        duparam = self._subelement(duplicate, "param")
        duparam.setAttribute("name", "index")
        duparam.setAttribute("use", ":" + name_id)


def to_sif(animation):
    builder = SifBuilder()
    builder.process(animation)
    return builder.dom
